from qs_mps.mps_class import MPS
import time
import numpy as np

# simulate different chi
L = 50
d = 2
model = "Ising"
chis = [16, 32, 64, 128]  # ,8,16,32,64,128,256,512
h = 1
J = 1
eps = 0
path = "/data/fdimarca/projects/0_ISING/results/time_data/"

time_tot = []
for chi in chis:
    print(f"start DMRG for L={L}, chi={chi}")
    time_st = time.perf_counter()
    chain = MPS(L=L, d=d, model=model, chi=chi, h=h, eps=eps, J=J)
    chain._random_state(seed=3, chi=chi)
    energy, entropy = chain.DMRG(trunc_chi=True, trunc_tol=False, n_sweeps=2)
    time_end = abs(time.perf_counter() - time_st)
    print(f"Time of computation: {time_end} sec")
    time_tot.append(time_end)

np.savetxt(f"{path}times_dmrg_L_{L}_chis_2-4-8-16-32-64-128", time_tot)

# load time

time_tot = np.loadtxt(f"{path}times_dmrg_L_{L}_chis_2-4-8-16-32-64-128")

# default parameters of the plot layout
import matplotlib.pyplot as plt
import matplotlib as mpl
from qs_mps.utils import fitting

plt.rcParams["text.usetex"] = True  # use latex
plt.rcParams["font.size"] = 13
plt.rcParams["figure.dpi"] = 300
plt.rcParams["figure.constrained_layout.use"] = True

import numpy as np
from scipy.optimize import curve_fit
import matplotlib.pyplot as plt


def exponential_function(x, a, b, c):
    """
    Exponential function: a * exp(b * x) + c
    """
    return a * np.exp(x * b) + c


def poly_function(x, a, b, c):
    """
    Poly function: a * x ** b + c
    """
    return a * x**b + c


def linear_function(x, a, b):
    """
    Linear function: a * x + b
    """
    return a * x + b


def fit(x_data, y_data, ftype):
    """
    Fit function to given data.

    Parameters:
    - x_data: Input data (independent variable)
    - y_data: Output data (dependent variable)
    - ftype: type of function we fit

    Returns:
    - Coefficients of the fitted function
    """
    # Use curve_fit to fit the data to the exponential function
    if ftype == "exp":
        params, covariance = curve_fit(exponential_function, x_data, y_data)
    if ftype == "poly":
        params, covariance = curve_fit(poly_function, x_data, y_data)
    if ftype == "lin":
        params, covariance = curve_fit(linear_function, x_data, y_data)

    # Extract the coefficients
    err = np.sqrt(np.diag(covariance))
    return params, err


# Example usage:
# Generate some example data
x_data = chis
# time_tot = np.array([3.23587389e+00,
#                     1.07527894e+01, 1.68719734e+02, 2.00160884e+03]) # 7.57573918e-02, 1.71739756e-01, 4.02117394e-01,
y_data = time_tot

# Fit the data to the exponential function
params_exp, err_exp = fit(x_data, y_data, "exp")
params_poly, err_poly = fit(x_data, y_data, "poly")

# Print the coefficients
print(f"Coefficients exp: {params_exp}")
print(f"Errors exp: {err_exp}")
print(f"Coefficients poly: {params_poly}")
print(f"Errors poly: {err_poly}")

# Plot the original data and the fitted curve
plt.title(f"DMRG Computational Time for $L={L}$ at $h=h_c$")
plt.scatter(x_data, y_data, label="Original Data")
x_fit = np.linspace(min(x_data), 1024, 1000)
# y_fit_exp = exponential_function(x_fit, params_exp[0], params_exp[1], params_exp[2])
y_fit_poly = poly_function(x_fit, params_poly[0], params_poly[1], params_poly[2])
# plt.plot(x_fit, y_fit_exp, 'r-', label='Fitted Exponential Curve')
plt.plot(x_fit, y_fit_poly, "g-", label="Fitted Polynomial Curve")
plt.fill_between(
    x_fit,
    poly_function(
        x_fit,
        params_poly[0] - err_poly[0],
        params_poly[1] - err_poly[1],
        params_poly[2] - err_poly[2],
    ),
    poly_function(
        x_fit,
        params_poly[0] + err_poly[0],
        params_poly[1] + err_poly[1],
        params_poly[2] + err_poly[2],
    ),
    color="green",
    alpha=0.2,
    label="Error Bounds",
)
plt.yscale("log")
# plt.xscale('log')
plt.xticks(ticks=chis + [1024], labels=chis + [1024])
plt.xlabel("bond dimension $(\chi)$")
plt.ylabel("time (t)")
y_last_plus = poly_function(
    x_fit[-1],
    params_poly[0] + err_poly[0],
    params_poly[1] + err_poly[1],
    params_poly[2] + err_poly[2],
)
print("last values")
print(y_fit_poly[-1])
print(y_last_plus)
y_err = abs(y_fit_poly[-1] - y_last_plus)

if y_fit_poly[-1] < 60:
    unit = "sec(s)"
elif y_fit_poly[-1] > 60 and y_fit_poly[-1] < 3600:
    unit = "min(s)"
    t = y_fit_poly[-1] / 60
    t_err = y_err / 60
elif y_fit_poly[-1] > 3600:
    unit = "hour(s)"
    t = y_fit_poly[-1] / 3600
    t_err = y_err / 3600

textstr = f"Expectation time for $\chi = 1024$ is {t:.2f} ± {t_err:.2f} {unit}"
plt.text(
    0.05,
    0.90,
    textstr,
    transform=plt.gca().transAxes,
    fontsize=10,
    verticalalignment="top",
    bbox=dict(boxstyle="round", facecolor="wheat", alpha=0.5),
)
textstr = f"Fitted Parameters:\na = {params_poly[0]:.2e} ± {err_poly[0]:.2e}\nb = {params_poly[1]:.2f} ± {err_poly[1]:.2f}\nc = {params_poly[2]:.2f} ± {err_poly[2]:.2f}"
plt.text(
    0.05,
    0.80,
    textstr,
    transform=plt.gca().transAxes,
    fontsize=10,
    verticalalignment="top",
    bbox=dict(boxstyle="round", facecolor="wheat", alpha=0.5),
)
textstr = f"Poly function:\n$a x^b + c$"
plt.text(
    0.33,
    0.80,
    textstr,
    transform=plt.gca().transAxes,
    fontsize=10,
    verticalalignment="top",
    bbox=dict(boxstyle="round", facecolor="wheat", alpha=0.5),
)
plt.hlines(
    y=y_fit_poly[-1],
    xmin=chis[0],
    xmax=1024,
    linestyles="dashed",
    colors="black",
    label='time $``=" \langle t \\rangle_{\chi=1024}$',
)
plt.hlines(
    y=3600,
    xmin=chis[0],
    xmax=1024,
    linestyles="dashed",
    colors="black",
    linewidth=0.8,
    label="one hour",
)
plt.legend(loc="lower right", fontsize=10)
plt.savefig(f"times_dmrg_L_{L}_chis_16-32-64-128.png")
plt.show()


# time for different L

Ls = [20, 30, 40, 50, 60, 70, 80]
d = 2
model = "Ising"
chi = 64  # ,8,16,32,64,128,256,512
# hs = np.linspace(0,2,100)
h = 1
J = 1
eps = 0
# time_tot = np.array([
#     4.262944635422900319e+01,
#     1.045260993340052664e+02,
#     1.592617676099762321e+02,
#     2.302356971697881818e+02,
#     3.002348961471579969e+02,
#     3.662781152101233602e+02,
#     4.503631118210032582e+02,
#     ])

x_data = Ls
y_data = time_tot

params_lin, err_lin = fit(x_data, y_data, "lin")

# Plot the original data and the fitted curve
plt.title(f"DMRG Computational Time for $\chi={chi}$ at $h=h_c$")
plt.scatter(x_data, y_data, label="Original Data")
x_fit = np.linspace(min(x_data), 1000, 1000)
y_fit_lin = linear_function(x_fit, params_lin[0], params_lin[1])
plt.plot(x_fit, y_fit_lin, "g-", label="Fitted Linear Curve")
plt.fill_between(
    x_fit,
    linear_function(x_fit, params_lin[0] - err_lin[0], params_lin[1] - err_lin[1]),
    linear_function(x_fit, params_lin[0] + err_lin[0], params_lin[1] + err_lin[1]),
    color="green",
    alpha=0.2,
    label="Error Bounds",
)
# plt.xscale('log')
plt.xticks(ticks=Ls + [1000], labels=Ls + [1000])
plt.xlabel("chain length $(L)$")
plt.ylabel("time (t)")
y_last_plus = linear_function(
    x_fit[-1], params_lin[0] + err_lin[0], params_lin[1] + err_lin[1]
)
print("last values")
print(y_fit_lin[-1])
print(y_last_plus)
y_err = abs(y_fit_lin[-1] - y_last_plus)

if y_fit_lin[-1] < 60:
    unit = "sec(s)"
elif y_fit_lin[-1] > 60 and y_fit_lin[-1] < 3600:
    unit = "min(s)"
    t = y_fit_lin[-1] / 60
    t_err = y_err / 60
elif y_fit_lin[-1] > 3600:
    unit = "hour(s)"
    t = y_fit_lin[-1] / 3600
    t_err = y_err / 3600

textstr = f"Expectation time for $L = 1000$ is {t:.2f} ± {t_err:.2f} {unit}"
plt.text(
    0.05,
    0.90,
    textstr,
    transform=plt.gca().transAxes,
    fontsize=10,
    verticalalignment="top",
    bbox=dict(boxstyle="round", facecolor="wheat", alpha=0.5),
)
textstr = f"Fitted Parameters:\na = {params_lin[0]:.2e} ± {err_lin[0]:.2e}\nb = {params_lin[1]:.2f} ± {err_lin[1]:.2f}"
plt.text(
    0.05,
    0.80,
    textstr,
    transform=plt.gca().transAxes,
    fontsize=10,
    verticalalignment="top",
    bbox=dict(boxstyle="round", facecolor="wheat", alpha=0.5),
)
textstr = f"lin function:\n$a x + b$"
plt.text(
    0.35,
    0.80,
    textstr,
    transform=plt.gca().transAxes,
    fontsize=10,
    verticalalignment="top",
    bbox=dict(boxstyle="round", facecolor="wheat", alpha=0.5),
)
plt.hlines(
    y=y_fit_lin[-1],
    xmin=chis[0],
    xmax=1024,
    linestyles="dashed",
    colors="black",
    label='time $``=" \langle t \\rangle_{L=1000}$',
)
plt.hlines(
    y=3600,
    xmin=chis[0],
    xmax=1024,
    linestyles="dashed",
    colors="black",
    linewidth=0.8,
    label="one hour",
)
plt.legend(loc="lower right", fontsize=10)
plt.savefig(f"times_dmrg_L_20-30-40-50-60-70-80_chi_{chi}.png")
plt.show()



# import packages
import argparse
from qs_mps.mps_class import MPS
from qs_mps.utils import *

parser = argparse.ArgumentParser(prog="observables_Ising_mps")
parser.add_argument(
    "npoints",
    help="Number of points in an interval of transverse field values",
    type=int,
)
parser.add_argument(
    "h_i", help="Starting value of h external transverse field", type=float
)
parser.add_argument(
    "h_f", help="Final value of h external transverse field", type=float
)
parser.add_argument(
    "path",
    help="Path to the drive depending on the device used. Available are 'pc', 'mac', 'marcos'",
    type=str,
)
parser.add_argument("o", help="Observable we want to compute. Available are 'mag'", type=str)
parser.add_argument("-L", "--Ls", help="Spin chain lengths", nargs="*", type=int)
parser.add_argument("-D", "--chis", help="Simulated bond dimensions", nargs="+", type=int)
parser.add_argument("-d", "--dimension", help="Physical dimension. By default 2", default=2, type=int)
parser.add_argument(
    "-m", "--model", help="Model to simulate", default="Ising", type=str
)
parser.add_argument(
    "-mo", "--moment", help="Moment degree of the Free energy. E.g. Magnetization -> First Moment, Susceptibility -> Second Moment, etc. Available are 1,2,4", default=1, type=int
)

args = parser.parse_args()

# define the interval of equally spaced values of external field
interval = np.linspace(args.h_i, args.h_f, args.npoints)

# take the path and precision to save files
# if we want to save the tensors we save them locally because they occupy a lot of memory
if args.path == "pc":
    parent_path = "G:/My Drive/projects/0_ISING"
    path_tensor = "D:/code/projects/0_ISING"
elif args.path == "mac":
    parent_path = "/Users/fradm98/Google Drive/My Drive/projects/0_ISING"
    path_tensor = "/Users/fradm98/Desktop/projects/0_ISING"
elif args.path == "marcos":
    parent_path = "/Users/fradm/Google Drive/My Drive/projects/0_ISING"
    path_tensor = "/Users/fradm/Desktop/projects/0_ISING"
else:
    raise SyntaxError("Path not valid. Choose among 'pc', 'mac', 'marcos'")

num = (args.h_f - args.h_i) / args.npoints
precision = get_precision(num) 


# define moment
if args.moment == 1:
    moment = "first"
if args.moment == 2:
    moment = "second"
if args.moment == 4:
    moment = "fourth"

# ---------------------------------------------------------
# Observables
# ---------------------------------------------------------
for L in args.Ls:
    for chi in args.chis:
        M = []
        for J in interval:
            for h in interval:
                chain_mps = MPS(L=L, d=args.dimension, model=args.model, chi=chi, h=h, J=J)

                chain_mps.load_sites(path=path_tensor, precision=precision)
                
                if args.o == "mag":
                    print(f"Magnetization for L:{L}, D:{chi}, h:{h:.{precision}f}, J:{J:.{precision}f}")
                    chain_mps.order_param()
                    if args.moment == 1:
                        M.append(chain_mps.mpo_first_moment().real)
                    elif args.moment == 2:
                        M.append(chain_mps.mpo_second_moment().real/(chain_mps.L**2))
                    elif args.moment == 4:
                        M.append(chain_mps.mpo_fourth_moment().real/(chain_mps.L**4))
                else:
                    raise ValueError("Select a valid observable. Available are 'mag'")

        if args.o == "mag":
            M = np.array(M)
            M = np.array_split(M, args.npoints)
            np.save(
                    f"{parent_path}/results/mag_data/magnetization_{moment}_moment_{args.model}_L_{L}_h_{args.h_i}-{args.h_f}_J_{args.h_i}-{args.h_f}_delta_{args.npoints}_chi_{chi}.npy",
                    M,
                )


import argparse
import numpy as np
from qs_mps.mps_class import MPS
from qs_mps.utils import get_precision, save_list_of_lists, access_txt, swap_columns, swap_rows
from qs_mps.applications.ANNNI.ground_state_multiprocessing import ground_state_ANNNI
from scipy.sparse.linalg._eigen.arpack.arpack import ArpackNoConvergence

# DENSITY MATRIX RENORMALIZATION GROUP to find ground states of the Z2 Pure Gauge Theory 
# changing the transverse field parameters in its dual formulation

parser = argparse.ArgumentParser(prog="gs_search_ANNNI")
parser.add_argument(
    "npoints",
    help="Number of points in an interval of transverse field values",
    type=int,
)
parser.add_argument(
    "hx_i", help="Starting value of h external transverse field", type=float
)
parser.add_argument(
    "hx_f", help="Final value of h external transverse field", type=float
)
parser.add_argument(
    "hy_i", help="Starting value of h external transverse field", type=float
)
parser.add_argument(
    "hy_f", help="Final value of h external transverse field", type=float
)
parser.add_argument(
    "path",
    help="Path to the drive depending on the device used. Available are 'pc', 'mac', 'marcos'",
    type=str,
)
parser.add_argument("-L", "--Ls", help="Spin chain lengths", nargs="*", type=int)
parser.add_argument("-D", "--chis", help="Simulated bond dimensions", nargs="+", type=int)
parser.add_argument("-d","--dimension", help="Physical dimension. By default 2", default=2, type=int)
parser.add_argument(
    "-ty", "--type_shape", help="Type of shape of the bond dimension. Available are: 'trapezoidal', 'pyramidal', 'rectangular'", default="rectangular", type=str
)
parser.add_argument(
    "-m", "--model", help="Model to simulate", default="ANNNI", type=str
)
parser.add_argument(
    "-mu", "--multpr", help="If True computes ground states with multiprocessing. By default False", action="store_true"
)
parser.add_argument(
    "-s",
    "--number_sweeps",
    help="Number of sweeps during the compression algorithm for each trotter step",
    default=8,
    type=int,
)
parser.add_argument(
    "-cv",
    "--conv_tol",
    help="Convergence tolerance of the compression algorithm",
    default=1e-10,
    type=float,
)
parser.add_argument(
    "-b",
    "--bond",
    help="Save the schmidt values for one bond. If False save for each bond. By default True",
    action="store_false",
)
parser.add_argument(
    "-w",
    "--where",
    help="Bond where we want to observe the Schmidt values, should be between 1 and (L-1)",
    default=-1,
    type=int,
)
parser.add_argument(
    "-v",
    "--save",
    help="Save the tensors. By default True",
    action="store_false",
)
parser.add_argument(
    "-tr",
    "--training",
    help="Save all the energies also the ones during the variational optimization. By default False",
    action="store_true",
)
parser.add_argument(
    "-i",
    "--interval",
    help="Type of interval spacing. Available are 'log', 'lin'",
    default="lin",
    type=str
)

args = parser.parse_args()

# define the physical dimension
d = int(2)

# define the interval of equally spaced values of external field
if args.interval == "lin":
    interval_k = np.linspace(args.hx_i, args.hx_f, args.npoints)
    interval_h = np.linspace(args.hy_i, args.hy_f, args.npoints)
    num = (args.hx_f - args.hx_i) / args.npoints
    precision = get_precision(num)
elif args.interval == "log":
    interval_k = np.logspace(args.hx_i, args.hx_f, args.npoints)
    interval_h = np.logspace(args.hy_i, args.hy_f, args.npoints)
    num = (interval_k[-1]-interval_k[0]) / args.npoints
    precision = get_precision(num)

interval_h = interval_h.tolist().copy()
interval_h.reverse()

# take the path and precision to save files
# if we want to save the tensors we save them locally because they occupy a lot of memory
if args.path == "pc":
    parent_path = "G:/My Drive/projects/2_ANNNI"
    path_tensor = "D:/code/projects/2_ANNNI"
elif args.path == "mac":
    parent_path = "/Users/fradm98/Google Drive/My Drive/projects/2_ANNNI"
    path_tensor = "/Users/fradm98/Desktop/projects/2_ANNNI"
elif args.path == "marcos":
    parent_path = "/Users/fradm/Google Drive/My Drive/projects/2_ANNNI"
    path_tensor = "/Users/fradm/Desktop/projects/2_ANNNI"
else:
    raise SyntaxError("Path not valid. Choose among 'pc', 'mac', 'marcos'")




# ---------------------------------------------------------
# DMRG
# ---------------------------------------------------------
for L in args.Ls:
    # where to look at for the entropy
    if args.where == -1:
        args.where = L // 2
    elif args.where == -2:
        args.bond = False
    for chi in args.chis:  # L // 2 + 1
        up = np.array([[[1],[0]]])
        init_tensor = [up for _ in range(L)]

        energy_chi = []
        entropy_chi = []
        schmidt_vals_chi = []
        for k in interval_k:
            energy_k = []
            entropy_k = []
            schmidt_vals_k = []
            for h in interval_h:
                chain = MPS(L=L, d=d, chi=chi, model=args.model, eps=1, h=h, J=1, k=k)
                chain.sites = init_tensor.copy()
                chain.enlarge_chi()
                energy, entropy, schmidt_vals = chain.DMRG(trunc_tol=False, trunc_chi=True, where=L//2)
                print(f"energy of h:{h:.{precision}f}, k:{k:.{precision}f} is:\n {energy}")
                print(f"Schmidt values in the middle of the chain:\n {schmidt_vals}")
                print(f"Entropy: {entropy}")
                
                if args.training:
                    energy_k.append(energy)
                else:
                    energy_k.append(energy[-1])
                entropy_k.append(entropy)
                schmidt_vals_k.append(schmidt_vals)
                chain.save_sites(path=path_tensor, precision=precision)
                if h == interval_h[0]:
                    init_tensor_k = chain.sites.copy()
                init_tensor = chain.sites.copy()
            
            init_tensor = init_tensor_k.copy()

            # energy_k.reverse()
            energy_chi.append(energy_k)
            entropy_chi.append(entropy_k)
            schmidt_vals_chi.append(schmidt_vals_k)

        if args.bond == False:
            args.where = "all"

        if args.training:
            energy_chi = swap_rows(np.asarray(energy_chi))
            # energy_chi = np.swapaxes(swap_columns(np.asarray(energy_chi)), axis1=0, axis2=1)
            # energy_chi = np.swapaxes(np.asarray(energy_chi), axis1=0, axis2=1)

            np.save(
                f"{parent_path}/results/energy_data/energies_{args.model}_L_{L}_k_{args.hx_i}-{args.hx_f}_h_{args.hy_i}-{args.hy_f}_delta_{args.npoints}_chi_{chi}",
                energy_chi,
            )
            energy_min = []
            for i in range(len(interval_k)):
                energy_min_i = []
                for j in range(len(interval_h)):
                    energy_min_i.append(energy_chi[i][j][-1])
                energy_min.append(energy_min_i)
            np.save(f"{parent_path}/results/energy_data/energy_{args.model}_L_{L}_k_{args.hx_i}-{args.hx_f}_h_{args.hy_i}-{args.hy_f}_delta_{args.npoints}_chi_{chi}", energy_min)
                    
        else:
            energy_chi = np.swapaxes(swap_columns(np.asarray(energy_chi)), axis1=0, axis2=1)
            np.save(
                f"{parent_path}/results/energy_data/energy_{args.model}_L_{L}_k_{args.hx_i}-{args.hx_f}_h_{args.hy_i}-{args.hy_f}_delta_{args.npoints}_chi_{chi}",
                energy_chi,
            )
        
        entropy_chi = np.swapaxes(swap_columns(np.asarray(entropy_chi), tensor=False), axis1=0, axis2=1)
        save_list_of_lists(
            f"{parent_path}/results/entropy_data/{args.where}_bond_entropy_{args.model}_L_{L}_k_{args.hx_i}-{args.hx_f}_h_{args.hy_i}-{args.hy_f}_delta_{args.npoints}_chi_{chi}",
            entropy_chi,
        )
        schmidt_vals_chi = np.swapaxes(swap_columns(np.asarray(schmidt_vals_chi)), axis1=0, axis2=1)
        save_list_of_lists(
            f"{parent_path}/results/entropy_data/{args.where}_schmidt_vals_{args.model}_L_{L}_k_{args.hx_i}-{args.hx_f}_h_{args.hy_i}-{args.hy_f}_delta_{args.npoints}_chi_{chi}",
            schmidt_vals_chi,
        )
        if args.where == "all":
            entropy_mid = access_txt(
                f"{parent_path}/results/entropy_data/{args.where}_bond_entropy_{args.model}_L_{L}_k_{args.hx_i}-{args.hx_f}_h_{args.hy_i}-{args.hy_f}_delta_{args.npoints}_chi_{chi}",
                (L-1) // 2,
            )
            np.savetxt(
                f"{parent_path}/results/entropy_data/{args.L // 2}_bond_entropy_{args.model}_L_{L}_k_{args.hx_i}-{args.hx_f}_h_{args.hy_i}-{args.hy_f}_delta_{args.npoints}_chi_{chi}",
                entropy_mid,
            )


import argparse
import json
import numpy as np
from qs_mps.utils import get_precision, save_list_of_lists, access_txt
from qs_mps.applications.Z2.ground_state_multiprocessing import ground_state_Z2
from qs_mps.mps_class import MPS
import os
import concurrent.futures

parser_ext = argparse.ArgumentParser(prog="gs_search_Z2_multi_run")
parser_ext.add_argument("p", help="path of the json files. Available paths are 'pc', 'mac', 'marcos'", type=str)
args_ext =parser_ext.parse_args()
# def parse_args():
#     parser = argparse.ArgumentParser(prog="gs_search_Z2")
#     parser.add_argument("l", help="Number of ladders in the direct lattice", type=int)
#     parser.add_argument("L", help="Number of rungs per ladder", type=int)
#     parser.add_argument(
#         "npoints",
#         help="Number of points in an interval of transverse field values",
#         type=int,
#     )
#     parser.add_argument(
#         "h_i", help="Starting value of h (external transverse field on the dual lattice)", type=float
#     )
#     parser.add_argument(
#         "h_f", help="Final value of h (external transverse field on the dual lattice)", type=float
#     )
#     parser.add_argument(
#         "path",
#         help="Path to the drive depending on the device used. Available are 'pc', 'mac', 'marcos'",
#         type=str,
#     )
#     parser.add_argument("chis", help="Simulated bond dimensions", nargs="+", type=int)
#     parser.add_argument("-cx", "--charges_x", help="a list of the first index of the charges", nargs="*", type=int)
#     parser.add_argument("-cy", "--charges_y", help="a list of the second index of the charges", nargs="*", type=int)

#     parser.add_argument(
#         "-ty", "--type_shape", help="Type of shape of the bond dimension. Available are: 'trapezoidal', 'pyramidal', 'rectangular'", default="trapezoidal", type=str
#     )
#     parser.add_argument(
#         "-m", "--model", help="Model to simulate", default="Z2_dual", type=str
#     )
#     parser.add_argument(
#         "-mu", "--multpr", help="If True computes ground states with multiprocessing. By default True", action="store_false"
#     )
#     parser.add_argument(
#         "-s",
#         "--number_sweeps",
#         help="Number of sweeps during the compression algorithm for each trotter step",
#         default=8,
#         type=int,
#     )
#     parser.add_argument(
#         "-cv",
#         "--conv_tol",
#         help="Convergence tolerance of the compression algorithm",
#         default=1e-10,
#         type=float,
#     )
#     parser.add_argument(
#         "-b",
#         "--bond",
#         help="Save the schmidt values for one bond. If False save for each bond. By default True",
#         action="store_false",
#     )
#     parser.add_argument(
#         "-w",
#         "--where",
#         help="Bond where we want to observe the Schmidt values, should be between 1 and (L-1)",
#         default=-1,
#         type=int,
#     )
#     parser.add_argument(
#         "-v",
#         "--save",
#         help="Save the tensors. By default True",
#         action="store_false",
#     )
#     parser.add_argument(
#         "-tr",
#         "--training",
#         help="Save all the energies also the ones during the variational optimization. By default True",
#         action="store_false",
#     )

#     return parser.parse_args()

# def save_config(run_number, args):
#     config = {
#         'run': run_number,
#         'arguments': vars(args)
#     }

#     with open(f'run_{run_number}_gs.json', 'w') as config_file:
#         json.dump(config, config_file, indent=2)

if args_ext.p == "pc":
    parent_path = "G:/My Drive/projects/1_Z2"
elif args_ext.p == "mac":
    parent_path = "/Users/fradm98/Google Drive/My Drive/projects/1_Z2"
elif args_ext.p == "marcos":
    parent_path = "/Users/fradm/Google Drive/My Drive/projects/1_Z2"
else:
    raise SyntaxError("Path not valid. Choose among 'pc', 'mac', 'marcos'")

def load_config(run_number):
    with open(f'{parent_path}/json_files/run_{run_number}_gs.json', 'r') as config_file:
        config = json.load(config_file)
        return argparse.Namespace(**config['arguments'])

def main():

    # Uncomment the next line if you want to resume from the last run
    # while os.path.exists(f'run_{run_number}.json'):
    for run_number in range(1,5):
        print(f"\nRun {run_number}:")
        

        # Save arguments to a configuration file
        # args = parse_args()
        # save_config(run_number, args)

        # Load arguments
        args = load_config(run_number)

        # Your main script logic here using the loaded arguments
        # define the physical dimension
        d = int(2**(args.l))

        # define the interval of equally spaced values of external field
        interval = np.linspace(args.h_i, args.h_f, args.npoints)

        # take the path and precision to save files
        # if we want to save the tensors we save them locally because they occupy a lot of memory
        if args.path == "pc":
            parent_path = "G:/My Drive/projects/1_Z2"
            path_tensor = "D:/code/projects/1_Z2"
        elif args.path == "mac":
            parent_path = "/Users/fradm98/Google Drive/My Drive/projects/1_Z2"
            path_tensor = "/Users/fradm98/Desktop/projects/1_Z2"
        elif args.path == "marcos":
            parent_path = "/Users/fradm/Google Drive/My Drive/projects/1_Z2"
            path_tensor = "/Users/fradm/Desktop/projects/1_Z2"
        else:
            raise SyntaxError("Path not valid. Choose among 'pc', 'mac', 'marcos'")

        num = (args.h_f - args.h_i) / args.npoints
        precision = get_precision(num)

        # define the sector by looking of the given charges
        # print(f"charge x: {type(args.charges_x)}")
        if args.charges_x == []:
            sector = "vacuum_sector"
            args.charges_x = None
            args.charges_y = None
        else:
            for i in range(1,args.l*args.L):
                if len(args.charges_x) == i:
                    sector = f"{i}_particle(s)_sector"

        # where to look at for the entropy
        if args.where == -1:
            args.where = args.L // 2
        elif args.where == -2:
            args.bond = False

        # ---------------------------------------------------------
        # DMRG
        # ---------------------------------------------------------
        for chi in args.chis:  # L // 2 + 1
            args_mps = {
                "L": args.L,
                "d": d,
                "chi": chi,
                "type_shape": args.type_shape,
                "model": args.model,
                "trunc_tol": False,
                "trunc_chi": True,
                "where": args.where,
                "bond": args.bond,
                "path": path_tensor,
                "save": args.save,
                "precision": precision,
                "sector": sector,
                "charges_x": args.charges_x,
                "charges_y": args.charges_y,
            }
            if __name__ == "__main__":
                energy_chi, entropy_chi = ground_state_Z2(
                    args_mps=args_mps, multpr=args.multpr, param=interval
                )

                # if args.bond == False:
                #     args.where = "all"

                # if args.training:
                #     save_list_of_lists(
                #         f"{parent_path}/results/energy_data/energies_{args.model}_direct_lattice_{args.l}x{args.L-1}_{sector}_{args.charges_x}-{args.charges_y}_h_{args.h_i}-{args.h_f}_delta_{args.npoints}_chi_{chi}",
                #         energy_chi,
                #     )
                #     energy_gs = access_txt(
                #             f"{parent_path}/results/energy_data/energies_{args.model}_direct_lattice_{args.l}x{args.L-1}_{sector}_{args.charges_x}-{args.charges_y}_h_{args.h_i}-{args.h_f}_delta_{args.npoints}_chi_{chi}",
                #             -1,
                #         )
                #     np.savetxt(
                #         f"{parent_path}/results/energy_data/energies_{args.model}_direct_lattice_{args.l}x{args.L-1}_{sector}_{args.charges_x}-{args.charges_y}_h_{args.h_i}-{args.h_f}_delta_{args.npoints}_chi_{chi}",
                #         energy_gs,
                #     )
                # else:
                #     np.savetxt(
                #         f"{parent_path}/results/energy_data/energies_{args.model}_direct_lattice_{args.l}x{args.L-1}_{sector}_{args.charges_x}-{args.charges_y}_h_{args.h_i}-{args.h_f}_delta_{args.npoints}_chi_{chi}",
                #         energy_chi,
                #     )
                    
                # save_list_of_lists(
                #     f"{parent_path}/results/entropy_data/{args.where}_bond_entropy_{args.model}_direct_lattice_{args.l}x{args.L-1}_{sector}_{args.charges_x}-{args.charges_y}_h_{args.h_i}-{args.h_f}_delta_{args.npoints}_chi_{chi}",
                #     entropy_chi,
                # )
                # if args.where == "all":
                #     entropy_mid = access_txt(
                #         f"{parent_path}/results/entropy_data/{args.where}_bond_entropy_{args.model}_direct_lattice_{args.l}x{args.L-1}_{sector}_{args.charges_x}-{args.charges_y}_h_{args.h_i}-{args.h_f}_delta_{args.npoints}_chi_{chi}",
                #         (args.L-1) // 2,
                #     )
                #     np.savetxt(
                #         f"{parent_path}/results/entropy_data/{args.L // 2}_bond_entropy_{args.model}_direct_lattice_{args.l}x{args.L-1}_{sector}_{args.charges_x}-{args.charges_y}_h_{args.h_i}-{args.h_f}_delta_{args.npoints}_chi_{chi}",
                #         entropy_mid,
                #     )

if __name__ == '__main__':
    main()



"""This module implements the Lattice base class for the LGT children classes"""

import numpy as np
from itertools import chain, product
from typing import Tuple, List, Dict

# Type hintings
Site = Tuple[int, int]
Plaquette = Tuple[int, int, int, int]
Star = Tuple[int, int, int, int]


#
def cyclic_perm(lst, n=1):
    return lst[n:] + lst[:n]

def zip_nearest(lst, periodic=True, step=1):
    if periodic:
        return zip(lst, cyclic_perm(lst, n=step))
    else:
        return zip(lst[:-step], lst[step:])


class LatticeError(RuntimeError):
    pass


class Lattice(object):
    def __init__(self, size: Tuple[int, int], pbc: Tuple[bool, bool]):
        """
        Create an instance of the Lattice class for the given size and
        periodic boundary conditions.
        This class is intended to be used as a base class for all the LGT models.

        Parameters
        ----------
        size : tuple(int, int)
            Number of sites along x and y respectively.
        pbc : tuple(bool, bool)
            Periodic boundary conditions along x and y respectively.
        """
        self.size = size
        self.pbc = pbc
        self._label_links()
        self._lattice_drawer = LatticeDrawer(self)

    @property
    def Lx(self) -> int:
        """Returns the horizontal length"""
        return self.size[0]

    @property
    def Ly(self) -> int:
        """Returns the vertical length"""
        return self.size[1]

    @property
    def pbc_x(self) -> bool:
        """Returns if x-axis is period"""
        return self.pbc[0]

    @property
    def pbc_y(self) -> bool:
        """Returns if y-axis is period"""
        return self.pbc[1]

    @property
    def nlinks_x(self) -> int:
        """Returns the number of links along the x-axis"""
        return self.Lx - 1 * (not self.pbc_x)

    @property
    def nlinks_y(self) -> int:
        """Returns the number of links along the y-axis"""
        return self.Ly - 1 * (not self.pbc_y)

    @property
    def area(self) -> int:
        """Returns the total area of the lattice"""
        return self.Lx * self.Ly

    @property
    def nlinks(self) -> int:
        """Returns the total number of links"""
        return self.Lx * self.nlinks_y + self.nlinks_x * self.Ly

    def __repr__(self):
        return f"<Lattice: size={self.size}, pbc={self.pbc}>"

    def __str__(self):
        ret = repr(self) + "\n\n"
        ret += self._lattice_drawer.draw_lattice()
        return ret

    def _label_links(self):
        """Labels the links of the lattice"""
        Lx = self.Lx
        Ly = self.Ly
        lx = self.nlinks_x
        ly = self.nlinks_y
        # initialize the "adjancy matrix"
        self._link_matrix = np.zeros((self.area, self.area))
        # assign for each link an integer in a sequential order
        # Starting from the horizontal ones and then the vertical ones
        # if there is no link than it's a zero
        if Lx == 1:  # first along x
            offset = 0  # nothing to do
        elif Lx == 2:
            y = np.arange(0, Ly)
            self._link_matrix[0 + y * Lx, 1 + y * Lx] = y + 1
            self._link_matrix[1 + y * Lx, 0 + y * Lx] = y + 1 + self.pbc_x * Ly
            offset = (1 + self.pbc_x) * Ly + 1
        else:
            x = np.arange(0, lx)
            for y in np.arange(0, Ly):
                self._link_matrix[x + y * Lx, ((x + 1) % Lx) + y * Lx] = x + y * lx + 1
                self._link_matrix[((x + 1) % Lx) + y * Lx, x + y * Lx] = x + y * lx + 1
                # ^ the link matrix have to be bidirectional
            offset = Ly * (Lx - 1 * (not self.pbc_x)) + 1

        if Ly == 2:  # then along y
            x = np.arange(0, Lx)
            self._link_matrix[x + 0 * Lx, x + 1 * Lx] = x + offset
            self._link_matrix[x + 1 * Lx, x + 0 * Lx] = (
                x + offset + Lx * int(self.pbc_y)
            )
        elif Ly > 2:
            y = np.arange(0, ly)
            for x in np.arange(0, Lx):
                self._link_matrix[x + y * Lx, x + ((y + 1) % Ly) * Lx] = (
                    x + y * Lx + offset
                )
                self._link_matrix[x + ((y + 1) % Ly) * Lx, x + y * Lx] = (
                    x + y * Lx + offset
                )

    def link(self, site0: Site, site1: Site, flip=False, from_zero=False) -> int:
        """Return the link index between two sites

        Parameters
        ----------
        site0 : tuple[int,int]
        site1 : tuple[int,int]
        flip : bool (default=False)
            whether to flip the integer repr of the link indices.
        from_zero : bool (default=False)
            whether to start the indexing from zero or not.
            May cause problems.

        the coordinates are taken mod Lx (for x_i) or mod Ly (for y_i)
        The value 0 (with from_zero=False) is used to indicate an absent link.

        Returns
        ----------
        The index of the link
        """
        x0, y0 = site0
        x1, y1 = site1
        index = (
            self._link_matrix[
                (x0 % self.Lx) + (y0 % self.Ly) * self.Lx,
                (x1 % self.Lx) + (y1 % self.Ly) * self.Lx,
            ]
            - 1 * from_zero
        )
        if flip:
            index = self.nlinks - index - 1 * from_zero
        return int(index)

    def plaquette(self, site: Site, **kwargs) -> Plaquette:
        """
        Return the links of a plaquette. The links are oriented
        counterclok-wise starting from the bottom site.

        Return
        ----------
        plaq : tuple[int, int, int, int]
            A 4-element list of the indices of the links of the plaquette.
        """
        x, y = site
        if x >= self.nlinks_x or y >= self.nlinks_y:
            raise ValueError(f"The coordinates {x} and {y} exceeds the lattice")

        plaq = (
            self.link((x, y), (x + 1, y), **kwargs),
            self.link((x + 1, y), (x + 1, y + 1), **kwargs),
            self.link((x + 1, y + 1), (x, y + 1), **kwargs),
            self.link((x, y + 1), (x, y), **kwargs),
        )
        return plaq

    def star(self, site: Site, L, l, **kwargs) -> Star:
        """
        Return the links of a star. The links are oriented counterclock-wise
        starting from the right link.

        Return
        ----------
        star : tuple[int, int, int, int]
            A 4-element array
        """
        x, y = site
        if x >= self.Lx or y >= self.Ly:
            raise ValueError(f"The coordinates {x} and {y} exceeds the lattice")

        star = []
        if x + 1 >= L:
            star.append(0)
        else:
            star.append(self.link((x, y), (x + 1, y), **kwargs))
        if y + 1 >= l:
            star.append(0)
        else:
            star.append(self.link((x, y), (x, y + 1), **kwargs))
        if x - 1 < 0:
            star.append(0)
        else:
            star.append(self.link((x - 1, y), (x, y), **kwargs))
        if y - 1 < 0:
            star.append(0)
        else:
            star.append(self.link((x, y - 1), (x, y), **kwargs))
        return star

    @property
    def sites(self) -> List[Site]:
        return list(product(range(self.Lx), range(self.Ly)))

    # @property
    def plaquettes(self, **kwargs) -> List[Plaquette]:
        """
        Return a list of all the plaquettes.
        The indices of the plaquettes start from 0 by default.
        """
        kwargs.setdefault("from_zero", False)
        plaquettes = [
            self.plaquette(site, **kwargs)
            for site in self.sites
            if site[0] < self.nlinks_x and site[1] < self.nlinks_y
        ]
        return plaquettes

    # @property
    def stars(self, **kwargs) -> List[Star]:
        """
        Return an array of all the stars.
        The indices of the stars start from 0.
        """
        kwargs.setdefault("from_zero", False)
        stars = [self.star(site, **kwargs) for site in self.sites]
        return stars

    def path(self, site0: Site, site1: Site, **kwargs) -> List[int]:
        """
        Return the indices of links that connect two sites that are on a straigth line

        Parameters
        ----------
        site0, site1 : tuple(int,int)
            the coordinates of the start and end of the path

        Return
        ----------
        path : np.array(int)
            A list of the links that connects site0 and site1.
            The indexing of the links starts from 0.

        Raise:
        ----------
        RuntimeError
                if the sites are not aligned
        """
        x0, y0 = site0
        x1, y1 = site1
        kwargs.setdefault("from_zero", False)
        step = lambda a, b: +1 if a <= b else -1
        if y0 == y1:
            sites = [(x, y0) for x in range(x0, x1 + step(x0, x1), step(x0, x1))]
        elif x0 == x1:
            sites = [(x0, y) for y in range(y0, y1 + step(y0, y1), step(y0, y1))]
        else:
            raise RuntimeError(
                f"The sites ({x0}, {y0}) and ({x1}, {y1}) are not aligned"
            )
        path_ = [
            self.link(prev, next, **kwargs)
            for prev, next in zip_nearest(sites, periodic=False)
        ]
        path_ = np.array(path_)
        return path_

    def dual_path(self, plq0: Site, plq1: Site, **kwargs) -> List[int]:
        """
        Given two plaquettes plq0 and plq1, it returns the links that are cut from
        a dual path that start from plq0 and ends in plq1.

        Parameters
        ----------
        plq0, plq1 : tuple(int,int)
            A plaquette is specified by the coordinates of the bottom-left site.

        Returns
        ----------
        dual_path : np.array(int)
            A list of indices of the links that are cut by the dual path.
            Indices start from 0.

        Raises
        ----------
        Exception
            If the plaquettes are not aligned along a straigth line
        """
        x0, y0 = plq0
        x1, y1 = plq1
        kwargs.setdefault("from_zero", False)
        # +1 for positive direction, -1 for negative directions:
        step = lambda a, b: +1 if a <= b else -1
        # get the correct starting and ending points
        # depending if we are going in the positive or negative direction
        begin = lambda a, b: a + 1 if a <= b else a
        end = lambda a, b: b + 1 if a <= b else a

        # is the path horizontal or vertical?
        if y0 == y1:
            pairs = [
                ((x, y0), (x, y0 + 1))
                for x in range(begin(x0, x1), end(x0, x1), step(x0, x1))
            ]
        elif x0 == x1:
            pairs = [
                ((x0, y), (x0 + 1, y))
                for y in range(begin(y0, y1), end(y0, y1), step(y0, y1))
            ]
        else:
            raise RuntimeError(
                f"The sites ({x0}, {y0}) and ({x1}, {y1}) are not aligned"
            )
        dual_path_ = [self.link(prev, next, **kwargs) for prev, next in pairs]
        dual_path_ = np.array(dual_path_)
        return dual_path_

    def loop(self, sites: List[Site], **kwargs):
        """
        Return a list of indices of a loop given a list of the site to visits.
        The sites are specified by tuple of two integers.

        Parameters
        ----------
        sites : list(tuple(int,int))

        Return
        ----------
        loop : np.array(int)
            A list of the indices of all the link visited
        """
        loop_indx = [
            self.path(prev, next, **kwargs)
            for prev, next in zip_nearest(sites, periodic=True)
        ]
        loop_indx = list(chain.from_iterable(loop_indx))
        return np.array(loop_indx)

    def string(self, plqs: List[Site], **kwargs):
        """
        Return a list of indices of a loop given a list of the site to visit.

        Parameters
        ----------
        plqs : list(tuple(int,int))

        Return
        ----------
        plaquettes : np.array(int)
            A list of the indices of all the link visited
        """
        string_indx = [
            self.dual_path(prev, next, **kwargs)
            for prev, next in zip_nearest(plqs, periodic=False)
        ]
        string_indx = list(chain.from_iterable(string_indx))
        return np.array(string_indx)

    def vlinks(self, y, **kwargs):
        """Returns a list of all the vertical links at a given y"""
        if y >= self.nlinks_y or y < 0:
            raise ValueError("Given y parameter exceeds the lattice dimensions")
        return [self.link((x, y), (x, y + 1), **kwargs) for x in range(self.Lx)]

    def hlinks(self, y, **kwargs):
        """Returns a list of all the horizontal links at a given y"""
        if y >= self.Ly or y < 0:
            raise ValueError("Given y parameter exceeds the lattice dimensions")
        bulk = [self.link((x, y), (x + 1, y), **kwargs) for x in range(self.Lx - 1)]
        if self.pbc_x:
            edges = [
                self.link((self.Lx - 1, y), (0, y), **kwargs),
                self.link((self.Lx - 1, y), (self.Lx, y), **kwargs),
            ]
        else:
            edges = []
        return (bulk, edges)


class LatticeDrawer(object):
    default_char_table = dict(hlink="--", vlink="|", site="+")

    def __init__(self, lattice, char_table: Dict = None):
        self.lattice = lattice
        self.Lx = self.lattice.Lx
        self.pbc_x = self.lattice.pbc_x
        self.char_table = (
            char_table if char_table is not None else self.default_char_table
        )

    def draw_row_links(self, y: int) -> str:
        """Draw the horizontal links of a given row"""
        hlink = self.char_table["hlink"]
        site = self.char_table["site"]
        blank = " " * len(hlink)
        bigblank = blank * 2 + "  "
        bulk, edges = self.lattice.hlinks(y, from_zero=False)
        left = blank + f"{edges[0]:>2d}" + hlink if self.pbc_x else bigblank
        right = hlink + f"{edges[1]:<2d}" + blank if self.pbc_x else bigblank
        left = left + site + hlink
        right = hlink + site + right
        inner = (hlink + site + hlink).join(f"{link:2d}" for link in bulk)
        return left + inner + right + "\n"

    def draw_vert_empty(self) -> str:
        """Draw empty vertical spaces"""
        blank = " " * (2 * len(self.char_table["hlink"]) + 2)
        vlink = self.char_table["vlink"]
        return vlink.join(blank for _ in range(self.Lx + 1)) + "\n"

    def draw_vert_links(self, y: int) -> str:
        """Draw the vertical links at a given y"""
        links = self.lattice.vlinks(y, from_zero=False)
        blank = " " * (2 * len(self.char_table["hlink"]) + 2)
        return blank + blank[1:].join(f"{link:<2d}" for link in links) + "\n"

    def draw_lattice(self) -> str:
        """Draw the lattice"""
        picture = ""
        for y in range(self.lattice.Ly - 1, 0, -1):
            picture += (
                self.draw_row_links(y)
                + self.draw_vert_empty()
                + self.draw_vert_links(y - 1)
                + self.draw_vert_empty()
            )
        picture += self.draw_row_links(0)

        if self.lattice.pbc_y:
            picture = (
                self.draw_vert_links(self.lattice.Ly - 1)
                + self.draw_vert_empty()
                + picture
                + self.draw_vert_empty()
                + self.draw_vert_links(self.lattice.Ly - 1)
            )
        return picture



import argparse
import json
import numpy as np
from qs_mps.utils import get_precision
from qs_mps.mps_class import MPS

parser_ext = argparse.ArgumentParser(prog="observables_Z2_multi_run")
parser_ext.add_argument("p", help="path of the json files. Available paths are 'pc', 'mac', 'marcos'", type=str)
args_ext =parser_ext.parse_args()

def parse_args():
    parser = argparse.ArgumentParser(prog="observables_Z2_mps")
    parser.add_argument("l", help="Number of ladders in the direct lattice", type=int)
    parser.add_argument("L", help="Number of rungs per ladder", type=int)
    parser.add_argument(
        "npoints",
        help="Number of points in an interval of transverse field values",
        type=int,
    )
    parser.add_argument(
        "h_i", help="Starting value of h (external transverse field on the dual lattice)", type=float
    )
    parser.add_argument(
        "h_f", help="Final value of h (external transverse field on the dual lattice)", type=float
    )
    parser.add_argument(
        "path",
        help="Path to the drive depending on the device used. Available are 'pc', 'mac', 'marcos'",
        type=str,
    )
    parser.add_argument("o", help="Observable we want to compute. Available are 'wl', 'el', 'thooft'", type=str)
    parser.add_argument("chis", help="Simulated bond dimensions", nargs="+", type=int)
    parser.add_argument("-cx", "--charges_x", help="a list of the first index of the charges", nargs="*", type=int)
    parser.add_argument("-cy", "--charges_y", help="a list of the second index of the charges", nargs="*", type=int)
    parser.add_argument("-s", "--sites", help="Number of sites in the wilson loop", nargs="*", type=int)
    parser.add_argument("-r", "--ladders", help="Number of ladders in the wilson loop", nargs="*", type=int)
    parser.add_argument(
        "-d", "--direction", help="Direction of the string", default="hor", type=str
    )
    parser.add_argument(
        "-m", "--model", help="Model to simulate", default="Z2_dual", type=str
    )

    return parser.parse_args()

def save_config(run_number, args):
    config = {
        'run': run_number,
        'arguments': vars(args)
    }

    with open(f'run_{run_number}_obs.json', 'w') as config_file:
        json.dump(config, config_file, indent=2)

if args_ext.p == "pc":
    parent_path = "G:/My Drive/projects/1_Z2"
elif args_ext.p == "mac":
    parent_path = "/Users/fradm98/Google Drive/My Drive/projects/1_Z2"
elif args_ext.p == "marcos":
    parent_path = "/Users/fradm/Google Drive/My Drive/projects/1_Z2"
else:
    raise SyntaxError("Path not valid. Choose among 'pc', 'mac', 'marcos'")

def load_config(run_number):
    with open(f'{parent_path}/json_files/run_{run_number}_obs.json', 'r') as config_file:
        config = json.load(config_file)
        return argparse.Namespace(**config['arguments'])

def main():

    # Uncomment the next line if you want to resume from the last run
    # while os.path.exists(f'run_{run_number}.json'):
    for run_number in range(1,5):
        print(f"\nRun {run_number}:")
        
        # Save arguments to a configuration file
        # args = parse_args()
        # save_config(run_number, args)

        # Load arguments
        args = load_config(run_number)

        # Your main script logic here using the loaded arguments
        # define the physical dimension
        d = int(2**(args.l))

        # define the interval of equally spaced values of external field
        interval = np.linspace(args.h_i, args.h_f, args.npoints)

        # take the path and precision to save files
        # if we want to save the tensors we save them locally because they occupy a lot of memory
        if args.path == "pc":
            parent_path = "G:/My Drive/projects/1_Z2"
            path_tensor = "D:/code/projects/1_Z2"
        elif args.path == "mac":
            parent_path = "/Users/fradm98/Google Drive/My Drive/projects/1_Z2"
            path_tensor = "/Users/fradm98/Desktop/projects/1_Z2"
        elif args.path == "marcos":
            parent_path = "/Users/fradm/Google Drive/My Drive/projects/1_Z2"
            path_tensor = "/Users/fradm/Desktop/projects/1_Z2"
        else:
            raise SyntaxError("Path not valid. Choose among 'pc', 'mac', 'marcos'")

        num = (args.h_f - args.h_i) / args.npoints
        precision = get_precision(num)

        # for the wilson loop
        if args.sites == 1:
            sites = 0
        if args.ladders == 1:
            ladders = 1

        # define the direction
        if args.direction == "ver":
            direction = "vertical"
        elif args.direction == "hor":
            direction = "horizontal"    

        # define the sector by looking of the given charges
        if len(args.charges_x) == 0:
            sector = "vacuum_sector"
            args.charges_x = None
            args.charges_y = None
        else:
            for i in range(1,args.l*args.L):
                if len(args.charges_x) == i:
                    sector = f"{i}_particle(s)_sector"

        # ---------------------------------------------------------
        # Observables
        # ---------------------------------------------------------
        for chi in args.chis:
            W = []
            E = []
            S = []
            for h in interval:
                lattice_mps = MPS(L=args.L, d=d, model=args.model, chi=chi, h=h)
                lattice_mps.L = lattice_mps.L - 1

                lattice_mps.load_sites(path=path_tensor, precision=precision, cx=args.charges_x, cy=args.charges_y)
                if sector != "vacuum_sector":
                    lattice_mps.Z2.add_charges(args.charges_x, args.charges_y)
                
                if args.o == "wl":
                    print(f"wilson loop for h:{h:.{precision}f}")
                    lattice_mps.Z2.wilson_Z2_dual(mpo_sites=[sites], ls=[ladders]) #list(range(s))
                    lattice_mps.w = lattice_mps.Z2.mpo.copy()
                    W.append(lattice_mps.mpo_first_moment().real)

                if args.o == "el":
                    print(f"electric field for h:{h:.{precision}f}")
                    E_h = np.zeros((2*args.l+1,2*args.L-1))
                    E_h[:] = np.nan
                    E_h = lattice_mps.electric_field_Z2(E_h)
                    E.append(E_h)
                
                if args.o == "thooft":
                    print(f"'t Hooft string for h:{h:.{precision}f}")
                    lattice_mps.Z2.thooft(site=args.sites, l=args.ladders, direction=direction)
                    lattice_mps.w = lattice_mps.Z2.mpo.copy()
                    S.append(lattice_mps.mpo_first_moment().real)


            if args.o == "wl":
                np.savetxt(
                            f"{parent_path}/results/wilson_loops/wilson_loop_{args.model}_direct_lattice_{args.l}x{args.L-1}_h_{args.h_i}-{args.h_f}_delta_{args.npoints}_chi_{chi}",
                            W,
                        )
            if args.o == "el":
                np.save(
                            f"{parent_path}/results/electric_field/electric_field_{args.model}_direct_lattice_{args.l}x{args.L-1}_{sector}_{args.charges_x}-{args.charges_y}_h_{args.h_i}-{args.h_f}_delta_{args.npoints}_chi_{chi}.npy",
                            E,
                        )
            if args.o == "hooft":
                np.save(
                            f"{parent_path}/results/thooft/thooft_string_{args.sites[0]}-{args.ladders[0]}_{direction}_{args.model}_direct_lattice_{args.l}x{args.L-1}_{sector}_{args.charges_x}-{args.charges_y}_h_{args.h_i}-{args.h_f}_delta_{args.npoints}_chi_{chi}.npy",
                            S,
                        )

if __name__ == '__main__':
    main()



import numpy as np
from qs_mps.utils import load_list_of_lists, tensor_shapes
from qs_mps.mps_class import MPS
import matplotlib.pyplot as plt

path_tensor = "/Users/fradm98/Desktop/projects/1_Z2"
charges_x = [[4,6],[3,7],[2,8],[1,9]]
charges_y = [2,2]
precision = 2
interval = np.logspace(-1,1,20)

# # compute the electric energy density for the vacuum
# eed_vacuum = []
# corr_vacuum = []
# for h in interval:
#     lattice_mps = MPS(L=11,d=2**4, model="Z2_dual", chi=256, h=h, J=1, w=[])
#     lattice_mps.L = lattice_mps.L - 1
#     lattice_mps.load_sites(path=path_tensor, precision=precision)
#     lattice_mps.connected_correlator(site=1, lad=2)
#     # eed = lattice_mps.electric_energy_density_Z2(site=4)
#     # eed_vacuum.append(eed)

# for cx in charges_x:
#     eed_charges = []
#     corr_charges = []
#     for h in interval:
#         lattice_mps = MPS(L=11,d=2**4, model="Z2_dual", chi=256, h=h, w=[])
#         lattice_mps.L = lattice_mps.L - 1
#         lattice_mps.Z2.add_charges(cx,charges_y)
#         lattice_mps.load_sites(path=path_tensor, precision=precision, cx=cx, cy=charges_y)
#         lattice_mps.connected_correlator(site=1, lad=2)
#         # eed = lattice_mps.electric_energy_density_Z2(site=4)
#         # eed_charges.append(eed)
    
#     # eed_conn = (eed_charges - eed_vacuum)
#     corr_conn = (np.asarray(corr_charges) - np.asarray(corr_vacuum))

#     # plt.title(f"Energy density for charges {cx}")
#     # plt.imshow(eed_conn)
#     # plt.show()
#     plt.title(f"Correlations for charges {cx}")
#     plt.imshow(corr_conn)
#     plt.show()

from matplotlib import colors
corr_vacuum = np.load("/Users/fradm98/Library/CloudStorage/GoogleDrive-fra.di.marcantonio@gmail.com/My Drive/projects/1_Z2/results/mag_data/connected_correlator_s_4_l_2_Z2_dual_direct_lattice_4x10_vacuum_sector_[]-[]_h_-1.0-1.0_delta_20_chi_256.npy")
corr_charges = np.load("/Users/fradm98/Library/CloudStorage/GoogleDrive-fra.di.marcantonio@gmail.com/My Drive/projects/1_Z2/results/mag_data/connected_correlator_s_4_l_2_Z2_dual_direct_lattice_4x10_2_particle(s)_sector_[2, 8]-[2, 2]_h_-1.0-1.0_delta_20_chi_256.npy")

conn_corr = corr_charges - corr_vacuum
print(corr_vacuum[0])
plt.imshow(conn_corr.T/2,aspect=15, cmap="seismic", interpolation="gaussian",  norm=colors.SymLogNorm(linthresh=1e-4,linscale=1e-4,
                                              vmin=-1.0, vmax=1.0, base=10))
plt.hlines(y=2+(np.log(2)/2),xmin=0,xmax=100,colors='gold',linestyles=':',linewidth=2)
plt.hlines(y=2-(np.log(2)/2),xmin=0,xmax=100,colors='gold',linestyles=':',linewidth=2)
# plt.imshow(corr_coeff.T/2,aspect=5, cmap="viridis", interpolation="gaussian", vmin=-1, vmax=1)
plt.colorbar()
plt.show()


# import packages
from qs_mps.utils import create_sequential_colors, tensor_shapes, mps_to_vector
from qs_mps.applications.Z2.ground_state_multiprocessing import ground_state_Z2
from qs_mps.applications.Z2.exact_hamiltonian import H_Z2_gauss
from qs_mps.sparse_hamiltonians_and_operators import sparse_pauli_x
from qs_mps.mps_class import MPS
import numpy as np
from scipy.sparse import identity
import matplotlib.pyplot as plt
import time


# finding the ground state of the vacuum sector
L = 5
l = 2
d = int(2**l)
chi = 16  # this is interpreted as d**(int(log2(chi))) --> e.g. chi=8 == 4**3=64
type_shape = "rectangular"
array = np.linspace(0.001, 10, 100)
hs = [h for h in array]
model = "Z2_dual"
path = "/Users/fradm98/Desktop/projects/1_Z2"
multpr = False
param = hs

if __name__ == "__main__":
    i = 0
    # for L in Ls:
    args_mps = {
        "L": L,
        "d": d,
        "chi": chi,
        "type_shape": type_shape,
        "model": model,
        "trunc_tol": False,
        "trunc_chi": True,
        "where": L // 2,
        "path": path,
        "save": True,
        "precision": 2,
        "sector": "vacuum_sector",
        "bond": True,
        "n_sweeps": 2,
        "conv_tol": 1e-10,
        "charges_x": None,
        "charges_y": None,
    }

    energies_h, entropy_h, schmidt_vals = ground_state_Z2(
        args_mps=args_mps, multpr=multpr, param=param
    )


# exact
eig_exact = []
eig_first = []
colors_mps = create_sequential_colors(20, "Blues")
colors = create_sequential_colors(20, "Reds")
markers = ["x","+","1","2"]

# alphas = [1,0.7]
l = 3
dof = (2*l*L - l - L)
print(dof)
v0 = np.array([-0.25 for _ in range(2**dof)])
for s in range(L-1):
    W = []
    W_exact = []
    for h in hs:
        print(f"h: {h}")
        ladder = MPS(L=L, d=d, model=model, h=h, eps=0, J=1, chi=chi)
        ladder.L = ladder.L - 1

        ladder.load_sites(path=path)
        ladder.Z2.wilson_Z2_dual(mpo_sites=[s], ls=[1]) #list(range(s))
        ladder.w = ladder.Z2.mpo
        W.append(ladder.mpo_first_moment().real)
    
        # exact
        Z2_exact = H_Z2_gauss(L=L, l=l, model="Z2", lamb=h, U=1e+3)
        H, e, v = Z2_exact.diagonalize(v0=v0)
        if s == 0:
            eig_exact.append(np.min(e))
        # eig_first.append(np.sort(e)[1])
        psi = v[:,0]
        v0 = psi
        # print("Psi exact:")
        # print(psi)

        loop = Z2_exact.latt.plaquettes(from_zero=True)
        plaq = Z2_exact.plaquette_term(loop[s])
        # plaq = Z2_exact.plaquette_term(loop[s+1])
        exp_val_wilson_loop = np.real(psi.T @ plaq @ psi)
        W_exact.append(exp_val_wilson_loop)


    # print(loop)
    plt.plot(hs, W, marker=markers[s], color="darkturquoise", label=f"mps {s}")
    plt.plot(hs, W_exact, markers[s], color="firebrick", label=f"exact {s}")


plt.xlabel("electric local parameter (h)")
plt.legend(loc="lower right")
plt.savefig("wilson_loop.png")
# plt.vlines(x=3.044, ymin=0, ymax=1)
# plt.title(f"Energy(h)")
# plt.xscale('symlog')
# plt.yscale('log')
# plt.xlabel("electric local parameter (h)")
# # plt.ylim([-350, -50])
# plt.legend(loc="lower right")
plt.figure().clear()
# plt.figure().clear()

fig = plt.figure()
plt.title(f"Energy(h)")
plt.scatter(hs,
            energies_h,
            marker="o",
            alpha=1,
            facecolors="none",
            edgecolors='g',
            label=f"mps L: {L}"
            )
plt.plot(hs,eig_exact,'--',color='red', label='exact gs')
plt.xlabel("electric local parameter (h)")
plt.legend(loc="lower right")
plt.savefig("energy.png")


print(ladder.Z2.latt._lattice_drawer.draw_lattice())


# print(ladder.Z2.latt._lattice_drawer.draw_lattice())

# plt.vlines(x=3.044, ymin=np.min(W), ymax=np.max(W))
# # plt.title(f"Energy(h)")
# # plt.yscale('log')
# plt.xlabel("electric local parameter (h)")
# # plt.ylim([-350, -50])
# plt.legend()
# plt.show()
# # %%
# plt.plot(hs, entropy_h)
# plt.vlines(x=3.044, ymin=np.min(entropy_h), ymax=np.max(entropy_h))


import concurrent.futures
import os
from .mps_class import MPS


def ground_state_ising_param(params):
    args_mps = params[0]
    param = params[1]
    precision = args_mps["precision"]
    chain = MPS(
        L=args_mps["L"],
        d=args_mps["d"],
        model=args_mps["model"],
        chi=args_mps["chi"],
        h=param,
        J=args_mps["J"],
        eps=args_mps["eps"],
    )
    chain._random_state(seed=7, chi=args_mps["chi"], type_shape=args_mps["type_shape"])
    chain.canonical_form(trunc_chi=False, trunc_tol=True)

    energy, entropy, schmidt_vals = chain.DMRG(
        trunc_tol=args_mps["trunc_tol"],
        trunc_chi=args_mps["trunc_chi"],
        where=args_mps["where"],
        bond=args_mps["bond"],
    )

    print(f"energy of h:{param:.{precision}f} is:\n {energy}")
    print(f"Schmidt values in the middle of the chain:\n {schmidt_vals}")

    chain.save_sites(path=args_mps["path"], precision=args_mps["precision"])
    return energy, entropy, schmidt_vals


def ground_state_ising_multpr(args_mps, multpr_param, cpu_percentage=90):
    max_workers = int(os.cpu_count() * (cpu_percentage / 100))
    with concurrent.futures.ProcessPoolExecutor(max_workers=max_workers) as executor:
        args = [[args_mps, param] for param in multpr_param]
        results = executor.map(ground_state_ising_param, args)

    energies = []
    entropies = []
    for result in results:
        energies.append(result[0])
        entropies.append(result[1])
    return energies, entropies


def ground_state_ising(args_mps, multpr, param):
    if multpr:
        energies_param, entropies_param = ground_state_ising_multpr(
            args_mps=args_mps, multpr_param=param
        )
    else:
        energies_param = []
        entropies_param = []
        schmidt_vals_param = []
        for p in param:
            params = [args_mps, p]
            energies, entropies, schmidt_vals = ground_state_ising_param(params=params)
            energies_param.append(energies)
            entropies_param.append(entropies)
            schmidt_vals_param.append(schmidt_vals)

    return energies_param, entropies_param, schmidt_vals_param



from qs_mps.lattice import Lattice
from qs_mps.sparse_hamiltonians_and_operators import sparse_pauli_z, sparse_pauli_x, sparse_pauli_y
from qs_mps.utils import mpo_to_matrix
from ncon import ncon
from scipy.sparse import csc_array, identity, linalg
import numpy as np


class MPO_ladder:
    def __init__(self, l, L, model=str, lamb=None):
        self.L = L
        self.l = l
        self.model = model
        self.charges = np.ones((l + 1, L))
        self.lamb = lamb
        self.mpo = []
        self.latt = self.latt = Lattice((self.L, self.l + 1), (False, False))
        self.dof = self.l * (self.L - 1)
        self.sector = self._define_sector()

    def charge_constraint(self):
        """
        charge_constraint

        We decide arbitrarily to impose a charge constraint on the bottom right
        corner of the direct lattice.

        """
        # take the product of the first l-1 rows
        productory = np.prod([self.charges[line, :] for line in range(self.l)])
        # take the last row but the last charge
        productory = productory * np.prod([self.charges[-1, :-1]])
        self.charges[-1, -1] = productory
        return self

    def add_charges(self, rows: list, columns: list):
        """
        add_charges

        This function adds the charges to the background
        vacuum sector (all positive charges). The number of
        charges we add are given by the len of each indices list

        rows: list - row indices of the charges to add
        columns: list - column indices of the charges to add
        """
        for i, j in zip(rows, columns):
            assert ((i != -1) and (j != -1)) or (
                (i == self.l - 1) and (j == self.L - 1)
            ), "Do not choose the last charge! We use it for Gauss Law constraint"
            self.charges[j, i] = -1
        # impose the constraint
        self.charge_constraint()
        self.charges = np.flip(self.charges, axis=1)
        return self

    def _define_sector(self):
        particles = 0
        for charge in self.charges.flatten():
            if charge == -1:
                particles += 1
        
        if particles == 0:
            sector = "vacuum_sector"
        else:
            sector = f"{particles}_sector"
        return sector
    
    def mpo_skeleton(self, aux_dim: int=None):
        """
        mpo_skeleton

        This function initializes the mpo tensor or shape (2+l,2+l,2**l,2**l)
        with O matrices. We add as well the identities in the first and last
        element of the mpo tensor.

        aux_dim: int - This auxiliary dimension represents how many rows and 
                    columns we want in our MPO. By default None means that it adapts
                    to the system under study. Fixing the auxiliary dimension is
                    useful for known observables which will not need larger MPOs
        """
        I = identity(2**self.l, dtype=complex)
        O = csc_array((2**self.l, 2**self.l), dtype=complex)
        if aux_dim == None:
            skeleton = np.array(
                [[O.toarray() for i in range(2 + self.l)] for j in range(2 + self.l)]
            )
        else:
            skeleton = np.array(
                [[O.toarray() for i in range(aux_dim)] for j in range(aux_dim)]
            )
        skeleton[0, 0] = I.toarray()
        skeleton[-1, -1] = I.toarray()
        self.mpo = skeleton
        return self

    def charge_coeff_interaction(self, n: int, mpo_site: int):
        """
        charge_coeff_interaction

        This function computes the coefficient for the interaction in the
        ladder Z2 hamiltonian.

        The coefficient

                c_n(j,j+1) = \\prod_{k=(n+1)}^{l+1} q_{k,(n+1)}

        is the product of the charges in the direct lattice "below"
        the interacting sites of the dual lattice.

        n: int - goes from 1 to l and indicate in which row of the dual lattice
            is happening the interaction
        mpo_site: int - indicates for which mpo we are computing the coefficient.
            There are (self.L-1) mpos and everyone encodes the interaction term with a
            specific charge coefficient. We take the charges starting from
            the row below the one of the interaction. Accepted values are from
            0 to (self.L - 2)

        """
        assert 1 <= n <= self.l, "Select a value of n within (1,l)"
        assert (
            0 <= mpo_site <= (self.L - 2)
        ), "Select a value of charge column within (0,L-2)"

        if mpo_site == (self.L - 2):
            c_n_j = 1
        else:
            c_n_j = np.prod(self.charges[n:, mpo_site + 1])
        return c_n_j

    def charge_coeff_local_Z(self, n: int, mpo_site: int):
        """
        charge_coeff_local_Z

        This function computes the coefficient for the local Z term in the
        ladder Z2 hamiltonian.

        """
        assert 1 <= n <= self.l, "Select a value of n within (1,l)"

        if mpo_site == 0 or mpo_site == (self.L - 2):
            if mpo_site == 0:
                col = mpo_site
                if n == 1:
                    alpha = 1
                    n = n - 1
                elif n == self.l:
                    alpha = 1
                    n = n
                else:
                    alpha = 0
            elif mpo_site == (self.L - 2):
                col = mpo_site + 1
                if n == 1:
                    alpha = 1
                    n = n - 1
                elif n == self.l:
                    alpha = 1
                    n = n
                else:
                    alpha = 0
                    

            c_n_j = (1 + self.charges[n, col]) ** (alpha) * np.prod(
                self.charges[n:, col]
            )
        else:
            if n == 1:
                c_n_j = np.prod([self.charges[:, col] for col in range(mpo_site + 1)])
            elif n == self.l:
                c_n_j = 1
            else:
                c_n_j = 0

        return c_n_j

    def charge_coeff_v_edge(self, mpo_site, l):
        """
        charge_coeff_v_edge

        This function gives the charge coefficients of the vertical edges of
        the fields in the direct lattice of a Z2 theory.

        """
        coeff = np.prod(self.charges[(l+1):, mpo_site])
        return coeff  

    def charge_coeff_v(self, mpo_site, l):
        """
        charge_coeff_v

        This function gives the charge coefficients of the vertical edges of
        the fields in the direct lattice of a Z2 theory.

        """
        if l == 0:
            coeff = np.prod([self.charges[:, col] for col in range(mpo_site + 1)])
        elif l == (self.l-1):
            coeff = 1
        return coeff
    
    def mpo_Z2_ladder_generalized(self):
        """
        mpo_Z2_ladder_generalized

        This function computes the hamiltonian MPO for the Z2
        gauge pure theory for general number of ladders and charge configuration.

        """
        self.mpo_skeleton()
        mpo_list = []
        for mpo_site in range(self.L - 1):
            # -----------
            # first row
            # -----------
            # interaction terms (from column 2 to l+1)
            for n in range(1, self.l + 1):
                # self.mpo[0,n] = - 1 * sparse_pauli_z(n=n-1, L=self.l).toarray()
                self.mpo[0, n] = (
                    -self.charge_coeff_interaction(n=n, mpo_site=mpo_site)
                    * self.lamb
                    * sparse_pauli_z(n=n - 1, L=self.l).toarray()
                )

            # local terms (column l+2)
            # local Z and local X
            for i in range(1, self.l + 1):
                # self.mpo[0,n+1] += - 3 * sparse_pauli_z(n=i-1, L=self.l).toarray() - self.lamb * sparse_pauli_x(n=i-1, L=self.l).toarray()
                self.mpo[0, n + 1] += (
                    -self.charge_coeff_local_Z(n=i, mpo_site=mpo_site)
                    * self.lamb
                    * sparse_pauli_z(n=i - 1, L=self.l).toarray()
                    - (1/self.lamb) * sparse_pauli_x(n=i - 1, L=self.l).toarray()
                )
            # vertical Z interaction
            for j in range(self.l - 1):
                self.mpo[0, n + 1] += -(
                    self.lamb
                    * sparse_pauli_z(n=j, L=self.l) @ sparse_pauli_z(n=j + 1, L=self.l)
                ).toarray()

            # -----------
            # last column
            # -----------
            # positive terms to complete the interaction between nearest neighbor sites (from row 2 to l+1)
            for n in range(1, self.l + 1):
                self.mpo[n, -1] = sparse_pauli_z(n=n - 1, L=self.l).toarray()

            mpo_list.append(self.mpo)
            self.mpo_skeleton()

        self.mpo = mpo_list
        return mpo_list

    def diagonalize(self):
        self.mpo_Z2_ladder_generalized()
        H = mpo_to_matrix(self.mpo)
        e, v = np.linalg.eigh(H)
        return e, v

    def _initialize_finalize_quench_local(self, delta, h_ev):
        I = identity(2**self.l, dtype=complex).toarray()
        w_tot = []
        for _ in range(self.L - 1):
            w_init_X = np.array([[I]])
            for l in range(self.l):
                X_l = sparse_pauli_x(n=l, L=self.l).toarray()
                w_init_X_l = np.array([[linalg.expm(1j * (1/h_ev) * delta * X_l)]])
                w_init_X = ncon([w_init_X_l, w_init_X], [[-1, -3, 1, -6], [-2, -4, -5, 1]]).reshape((1,1,w_init_X.shape[2],w_init_X_l.shape[3]))
            w_tot.append(w_init_X)
        self.mpo = w_tot
        return self

    def mpo_Z2_quench_tot(self, delta, h_ev):
        I = identity(2**self.l, dtype=complex).toarray()

        w_tot = []
        for mpo_site in range(self.L - 1):   
            w_l = np.array([[I]])
            for l in range(1,self.l+1):
                Z = sparse_pauli_z(n=l - 1, L=self.l).toarray()
                c_loc = self.charge_coeff_local_Z(n=l, mpo_site=mpo_site)
                w_loc = np.array(linalg.expm(1j * c_loc * h_ev * delta * Z))

                c_int = self.charge_coeff_interaction(n=l, mpo_site=mpo_site)
                w_even = np.array(
                    [
                        [
                            np.sqrt(np.cos(c_int * h_ev * delta)) * I,
                            1j * np.sign(c_int) * np.sqrt(np.sin(np.abs(c_int) * h_ev * delta)) * Z,
                        ]
                    ]
                )
                w_odd = np.array(
                    [
                        [
                            np.sqrt(np.cos(c_int * h_ev * delta)) * I,
                            np.sqrt(np.sin(np.abs(c_int) * h_ev * delta)) * Z,
                        ]
                    ]
                )
                w_odd = np.swapaxes(w_odd, axis1=0, axis2=1)
                if mpo_site == 0:
                    w_l = ncon(
                        [w_loc, w_even, w_l],
                        [[-5, 2], [-1, -3, 2, 1], [-2, -4, 1, -6]],
                    ).reshape((w_l.shape[0]*w_even.shape[0],w_l.shape[1]*w_even.shape[1],w_loc.shape[0],w_l.shape[-1]))
                elif mpo_site == self.L - 2:
                    w_l = ncon(
                        [w_loc,w_odd,w_l],
                        [[-5, 2], [-1, -3, 2, 1], [-2, -4, 1, -6]],
                    ).reshape((w_l.shape[0]*w_odd.shape[0],w_l.shape[1]*w_odd.shape[1],w_loc.shape[0],w_l.shape[-1]))
                else:
                    w_l = ncon(
                        [w_loc, w_even, w_odd, w_l],
                        [[-7, 3], [-1, -4, 3, 2], [-2, -5, 2, 1], [-3, -6, 1, -8]],
                    ).reshape((w_l.shape[0]*w_odd.shape[0]*w_even.shape[0],w_l.shape[1]*w_odd.shape[1]*w_even.shape[1],w_loc.shape[0],w_l.shape[-1]))

            for l in range(1,self.l):
                Z_ll = (
                    sparse_pauli_z(n=l - 1, L=self.l) @ sparse_pauli_z(n=l, L=self.l)
                ).toarray()
                w_int_loc = np.array(linalg.expm(1j * h_ev * delta * Z_ll))
                w_l = ncon([w_int_loc,w_l],[[-3,1],[-1,-2,1,-4]])

            w_tot.append(w_l)

        self.mpo = w_tot
        return self
    
    def thooft(self, site: list, l: list, direction: str):
        """
        thooft

        This function finds the 't Hooft string for the Z2 dual lattice.
        It gives us vertical and horizontal strings from a specific dual lattice
        site and going, conventionally, up and left, respectively (to vertical and horizontal).
        
        site: list - the first element is the mps site of the interested dual lattice site, starts from 0
        l: list - the first element is the ladder of the interested dual lattice site, starts from 0
        direction: str - indicates in the direction of the string. We use the convention:
                hor -> from left to 'site' at a specific 'l'
                ver -> from up to 'l' at a specific 'site'

        """
        self.mpo_skeleton(aux_dim=2)

        site = site[0]
        l = l[0]
        mpo_tot = []
        if direction == "vertical":
            coeff = self.charge_coeff_v(mpo_site=site, l=0)
        if direction == "horizontal":
            coeff = self.charge_coeff_v_edge(mpo_site=site, l=l)
            for mpo_site in range(site):
                coeff = coeff * self.charge_coeff_interaction(n=l+1, mpo_site=mpo_site)

        for mpo_site in range(self.L-1):
            if mpo_site == site:
                self.mpo[0,-1] = coeff * sparse_pauli_z(n=l, L=self.l).toarray()
            mpo_tot.append(self.mpo)
            self.mpo_skeleton(aux_dim=2)
                    
        self.mpo = mpo_tot
        return self
    
    def correlator(self, site: list, ladders: list):
        """
        thooft

        This function finds the 't Hooft string for the Z2 dual lattice.
        It gives us vertical and horizontal strings from a specific dual lattice
        site and going, conventionally, up and left, respectively (to vertical and horizontal).
        
        site: list - the first element is the mps site of the interested dual lattice site
        l: list - the first element is the ladder of the interested dual lattice site
        direction: str - indicates in the direction of the string. We use the convention:
                hor -> from left to 'site' at a specific 'l'
                ver -> from up to 'l' at a specific 'site'

        """
        self.mpo_skeleton(aux_dim=2)

        site = site[0]
        mpo_tot = []
        
        for mpo_site in range(self.L-1):
            if mpo_site == site:
                self.mpo[0,-1] = identity(2**self.l, dtype=complex).toarray()
                for l in ladders:
                    if l == 0 or l == self.l:
                        if l == 0:
                            l = l
                        elif l == self.l:
                            l = l - 1 
                        coeff = self.charge_coeff_v(mpo_site=site, l=l)
                        self.mpo[0,-1] = self.mpo[0,-1] @ (coeff * sparse_pauli_z(n=l, L=self.l).toarray())
                        # self.mpo[0,-1] = self.mpo[0,-1] @ (sparse_pauli_z(n=l, L=self.l).toarray())
                    else:
                        l = l - 1
                        self.mpo[0,-1] = self.mpo[0,-1] @ sparse_pauli_z(n=l, L=self.l).toarray() @ sparse_pauli_z(n=l+1, L=self.l).toarray()
            mpo_tot.append(self.mpo)
            self.mpo_skeleton(aux_dim=2)
                    
        self.mpo = mpo_tot
        return self

    def wilson_Z2_dual(self, mpo_sites: list, ls: list):
        """
        wilson_Z2_dual

        This function finds the wilson loop given by a list of indices in the lattice.

        mpo_sites: list - define the x axis of the lattice, starts from zero (left). n° of rungs
        ls: list - define the y axis of the lattice, starts from 1 (up). n° of ladders

        """
        # initialize
        self.mpo_skeleton()
        I = identity(2**self.l, dtype=complex).toarray()
        mpo_tot = []

        for site in range(self.L - 1):
            if len(mpo_sites) == 1:
                # we are between two rungs
                if site in mpo_sites and (mpo_sites[0] == 0):
                    self.mpo[0, 1] = self.pauli_string(string=ls, direction="vertical", pauli_type="X")
                elif site in mpo_sites:
                    self.mpo[1, -1] = self.pauli_string(string=ls, direction="vertical", pauli_type="X")
            # take the string of pauli on the ladders that creates the interaction among the mpo sites
            else:
                if site == mpo_sites[0]:
                    self.mpo[0, 1] = self.pauli_string(string=ls, direction="vertical", pauli_type="X")
                # take the string of pauli on the ladders that ends the interaction among the mpo sites
                elif site == mpo_sites[-1]:
                    self.mpo[1, -1] = self.pauli_string(string=ls, direction="vertical", pauli_type="X")
                # take the string of pauli on the ladders that continues the interaction among the mpo sites
                elif site in mpo_sites[1:-1]:
                    self.mpo[1, 1] = self.pauli_string(string=ls, direction="vertical", pauli_type="X")
            
            if site < mpo_sites[0]:
                # before the pauli strings, place identities for the previous mpo sites
                self.mpo[0, 1] = I
            if site > mpo_sites[-1]:
                # after the pauli strings, place identities for the remaining mpo sites
                self.mpo[1, -1] = I

            mpo_tot.append(self.mpo)
            self.mpo_skeleton()

        self.mpo = mpo_tot
        return self

    def pauli_string(self, string: list, direction: str, pauli_type: str):
        """
        pauli_string

        This function creates a string of certain paulis in a given direction 
        of the lattice.

        string: list - list of the sites on the dual lattice. Starts from 1
        direction: str - direction of the string on the sites of the dual lattice
        pauli_type: str - type of pauli matrix we want to use. Available are: "X", "Y", "Z"
        
        """
        # define the pauli type
        if pauli_type == "X":
            sigma = sparse_pauli_x
        elif pauli_type == "Y":
            sigma = sparse_pauli_y
        elif pauli_type == "Z":
            sigma = sparse_pauli_z
        
        # define the direction
        if direction == "horizontal":
            dim = self.L-1
        elif direction == "vertical":
            dim = self.l

        # compute the string
        pauli = identity(n=2**dim, dtype=complex)
        for n in string:
            pauli = pauli @ sigma(n=n, L=dim)

        return pauli.toarray()

    def local_observable_Z2_dual(self, mpo_site, l):
        """
        local_observable_Z2_dual

        This function finds the mpo representing local observables in the 
        dual lattice of the Z2 theory.

        """
        assert (0 <= mpo_site < self.L-1), "The mpo site is out of bound. Choose it 0 <= site < L-1"
        assert (0 <= l < self.l), "The mpo ladder is out of bound. Choose it 0 <= site < l"

        self.mpo_skeleton(aux_dim=2)

        mpo_tot = []
        for site in range(self.L - 1):
            if mpo_site == site:
                self.mpo[0, -1] = sparse_pauli_z(n=l, L=self.l).toarray()

            mpo_tot.append(self.mpo)
            self.mpo_skeleton(aux_dim=2)

        self.mpo = mpo_tot
        return self

    def zz_observable_Z2_dual(self, mpo_site: int, l: int, direction: str, aux_dim: int=2):
        """
        zz_observable_Z2_dual

        This function finds the mpo representing the interacting observales in the
        dual lattice of the Z2 theory. We make a distinction in the direction of
        the interaction because in our mps formalism the vertical interaction falls
        back into a local one.

        """
        self.mpo_skeleton(aux_dim=aux_dim)

        mpo_tot = []
        i = 0

        for site in range(self.L - 1):
            if direction == "horizontal":
                assert (0 <= mpo_site < self.L-2), "The mpo site is out of bound. Choose it 0 <= site < L-2"
                if (mpo_site+i) == site:
                    if i == 0:
                        self.mpo[0, 1] = sparse_pauli_z(n=l, L=self.l).toarray()
                    else:
                        self.mpo[1, -1] = sparse_pauli_z(n=l, L=self.l).toarray()
                    i = 1
            elif direction == "vertical":
                assert (0 <= l < self.l-1), "The mpo ladder is out of bound. Choose it 0 <= site < l-1"
                if mpo_site == site:
                    self.mpo[0, -1] = (
                        sparse_pauli_z(n=l, L=self.l).toarray()
                        @ sparse_pauli_z(n=l + 1, L=self.l).toarray()
                    )

            mpo_tot.append(self.mpo)
            self.mpo_skeleton(aux_dim=aux_dim)

        self.mpo = mpo_tot
        return self



# import packages
import argparse
from qs_mps.applications.Z2.exact_hamiltonian import H_Z2_gauss
from qs_mps.utils import *
import matplotlib.pyplot as plt


parser = argparse.ArgumentParser(prog="observables_Z2_exact")
parser.add_argument("l", help="Number of ladders in the direct lattice", type=int)
parser.add_argument("L", help="Number of rungs per ladder", type=int)
parser.add_argument(
    "npoints",
    help="Number of points in an interval of transverse field values",
    type=int,
)
parser.add_argument(
    "h_i", help="Starting value of h (external transverse field on the dual lattice)", type=float
)
parser.add_argument(
    "h_f", help="Final value of h (external transverse field on the dual lattice)", type=float
)
parser.add_argument(
    "path",
    help="Path to the drive depending on the device used. Available are 'pc', 'mac', 'marcos'",
    type=str,
)
parser.add_argument("o", help="Observable we want to compute. Available are 'wl', 'el', 'thooft'", type=str)
parser.add_argument("-cx", "--charges_x", help="a list of the first index of the charges", nargs="*", type=int)
parser.add_argument("-cy", "--charges_y", help="a list of the second index of the charges", nargs="*", type=int)
parser.add_argument("-s","--sites", help="Indices of sites in the wilson loop. Start from 0, left", nargs="*", type=int)
parser.add_argument("-r","--ladders", help="Indices of ladders in the wilson loop. Start from 1, above", nargs="*", type=int)
parser.add_argument(
    "-d", "--direction", help="Direction of the string", default="hor", type=str
)
parser.add_argument(
    "-m", "--model", help="Model to simulate. By default Z2_dual", default="Z2_dual", type=str
)
parser.add_argument(
    "-U", "--gauss", help="Gauss constraint parameter. By default 1e+3", default=1e+3, type=float
)

args = parser.parse_args()

# define the interval of equally spaced values of external field
interval = np.linspace(args.h_i, args.h_f, args.npoints)

# take the path and precision to save files
# if we want to save the tensors we save them locally because they occupy a lot of memory
if args.path == "pc":
    parent_path = "G:/My Drive/projects/1_Z2"
    path_eigvec = "D:/code/projects/1_Z2"
elif args.path == "mac":
    parent_path = "/Users/fradm98/Google Drive/My Drive/projects/1_Z2"
    path_eigvec = "/Users/fradm98/Desktop/projects/1_Z2"
elif args.path == "marcos":
    parent_path = "/Users/fradm/Google Drive/My Drive/projects/1_Z2"
    path_eigvec = "/Users/fradm/Desktop/projects/1_Z2"
else:
    raise SyntaxError("Path not valid. Choose among 'pc', 'mac', 'marcos'")

num = (args.h_f - args.h_i) / args.npoints
precision = get_precision(num)

# define the direction
if args.direction == "ver":
    direction = "vertical"
elif args.direction == "hor":
    direction = "horizontal"   

# define the sector by looking of the given charges
if len(args.charges_x) == 0:
    sector = "vacuum_sector"
    args.charges_x = None
    args.charges_y = None
else:
    for i in range(1,args.l*args.L):
        if len(args.charges_x) == i:
            sector = f"{i}_particle(s)_sector"

# ---------------------------------------------------------
# Observables
# ---------------------------------------------------------
W = []
E = []
S = []

for h in interval:

    Z2_exact = H_Z2_gauss(L=args.L, l=args.l, model=args.model, lamb=h, U=1e+3)
    # print(Z2_exact.latt._lattice_drawer.draw_lattice())
    psi = np.load(f"{path_eigvec}/results/eigenvectors/ground_state_direct_lattice_{args.l-1}x{args.L-1}_{sector}_{args.charges_x}-{args.charges_y}_U_{args.gauss}_h_{h:.{precision}f}.npy")
    # print(psi)

    if args.o == "wl":
        W.append(Z2_exact.wilson_loop(psi, args.sites, args.ladders))
    
    if args.o == "el":
        print(f"electric field for h:{h:.{precision}f}")
        E_h = np.zeros((2*args.l-1,2*args.L-1))
        E_h[:] = np.nan
        E_h = Z2_exact.electric_field(psi, E_h)
        # print(E_h)
        E.append(E_h)

    if args.o == "thooft":
        print(f"'t Hooft string for h:{h:.{precision}f}")
        s_h = Z2_exact.thooft(psi=psi, mpo_site=args.sites[0], l=args.ladders[0], direction=direction)
        S.append(s_h)

if args.o == "wl":
    np.savetxt(
                f"{parent_path}/results/exact/wilson_loops/wilson_loop_{args.model}_direct_lattice_{args.l-1}x{args.L-1}_{sector}_h_{args.h_i}-{args.h_f}_delta_{args.npoints}",
                W,
            )
if args.o == "el":
    np.save(
                f"{parent_path}/results/exact/electric_field/electric_field_{args.model}_direct_lattice_{args.l-1}x{args.L-1}_{sector}_{args.charges_x}-{args.charges_y}_h_{args.h_i}-{args.h_f}_delta_{args.npoints}.npy",
                E,
            )
if args.o == "thooft":
    np.save(
                f"{parent_path}/results/exact/thooft/thooft_string_{args.sites[0]}-{args.ladders[0]}_{direction}_{args.model}_direct_lattice_{args.l-1}x{args.L-1}_{sector}_{args.charges_x}-{args.charges_y}_h_{args.h_i}-{args.h_f}_delta_{args.npoints}.npy",
                S,
            )
    
    


import matplotlib.pyplot as plt
import numpy as np
from qs_mps.utils import load_list_of_lists, tensor_shapes
from qs_mps.mps_class import MPS
from scipy.optimize import curve_fit

# TIME EVOLVING BLOCK DECIMATION to find evolutions of ground states of the Z2 Pure Gauge Theory 
# changing the transverse field parameters in its dual formulation for the Quench Hamiltonian

import argparse
from qs_mps.mps_class import MPS
from qs_mps.utils import *

parser = argparse.ArgumentParser(prog="Time Ev")
parser.add_argument("l", help="Number of ladders in the direct lattice", type=int)
parser.add_argument(
    "npoints",
    help="Trotter steps for the quench dynamics",
    type=int,
)
parser.add_argument(
    "delta", help="Width of each time slice during the time evolution. Should be 'small enough'", type=float
)
parser.add_argument(
    "h_i", help="Starting value of h (external transverse field on the dual lattice)", type=float
)
parser.add_argument(
    "h_ev", help="Quench value of h (external transverse field on the dual lattice)", type=float
)
parser.add_argument(
    "path",
    help="Path to the drive depending on the device used. Available are 'pc', 'mac', 'marcos'",
    type=str,
)
parser.add_argument("-L", "--Ls", help="Number of rungs per ladder", nargs="+", type=int)
parser.add_argument("-D", "--chis", help="Simulated bond dimensions", nargs="+", type=int)
parser.add_argument("-o", "--obs", help="Observable we want to compute. Available are 'wl', 'el', 'thooft', 'mag', 'eed'", nargs="*", type=str)
parser.add_argument("-cx", "--charges_x", help="a list of the first index of the charges", nargs="*", type=int)
parser.add_argument("-cy", "--charges_y", help="a list of the second index of the charges", nargs="*", type=int)
parser.add_argument("-lx", "--sites", help="Number of sites in the wilson loop", nargs="*", type=int)
parser.add_argument("-ly", "--ladders", help="Number of ladders in the wilson loop", nargs="*", type=int)
parser.add_argument(
    "-d", "--direction", help="Direction of the string", default="hor", type=str
)
parser.add_argument(
    "-b",
    "--bond",
    help="Save the schmidt values for one bond. If False save for each bond. By default True",
    action="store_false",
)
parser.add_argument(
    "-w",
    "--where",
    help="Bond where we want to observe the Schmidt values, should be between 1 and (L-1)",
    default=-1,
    type=int,
)
parser.add_argument(
    "-m", "--model", help="Model to simulate", default="Z2_dual", type=str
)
parser.add_argument(
    "-mo", "--moment", help="Moment degree of the Free energy. E.g. Magnetization -> First Moment, Susceptibility -> Second Moment, etc. Available are 1,2,4", default=1, type=int
)
parser.add_argument(
    "-tr",
    "--training",
    help="Save all the errors during the variational compression. By default False",
    action="store_true",
)

args = parser.parse_args()

# define the physical dimension
d = int(2**(args.l))

# define the precision to load the mps
# precision = get_precision(args.h_i)
precision = 2

# take the path and precision to save files
# if we want to save the tensors we save them locally because they occupy a lot of memory
if args.path == "pc":
    parent_path = "G:/My Drive/projects/1_Z2"
    path_tensor = "D:/code/projects/1_Z2"
elif args.path == "mac":
    parent_path = "/Users/fradm98/Google Drive/My Drive/projects/1_Z2"
    path_tensor = "/Users/fradm98/Desktop/projects/1_Z2"
elif args.path == "marcos":
    parent_path = "/Users/fradm/Google Drive/My Drive/projects/1_Z2"
    path_tensor = "/Users/fradm/Desktop/projects/1_Z2"
else:
    raise SyntaxError("Path not valid. Choose among 'pc', 'mac', 'marcos'")


# for the wilson loop
if args.sites == 1:
    sites = 0
if args.ladders == 1:
    ladders = 1

# define the direction
if args.direction == "ver":
    direction = "vertical"
elif args.direction == "hor":
    direction = "horizontal"    

# define moment
if args.moment == 1:
    moment = "first"
if args.moment == 2:
    moment = "second"
if args.moment == 4:
    moment = "fourth"

# define all observables
if args.obs == []:
    args.obs = ["wl","el","thooft","mag"]


for L in args.Ls:
    # define the sector by looking of the given charges
    if len(args.charges_x) == 0:
        sector = "vacuum_sector"
        charges_x = None
        charges_y = None
    else:
        sector = f"{len(args.charges_x)}_particle(s)_sector"
        charges_x = args.charges_x
        charges_y = args.charges_y

    # where to look at for the entropy
    if args.where == -1:
        args.where = L // 2
    elif args.where == -2:
        args.bond = False

    for chi in args.chis:
        W = []
        El = []
        S = []
        M = []
        C = []
        Ed = []

        # initialize the tensor
        lattice_mps = MPS(L=L, d=d, model=args.model, chi=chi, h=args.h_i)
        lattice_mps.L = lattice_mps.L - 1
        
        # load the mps
        lattice_mps.load_sites(path=path_tensor, precision=precision, cx=charges_x, cy=charges_y)
        if sector != "vacuum_sector":
            lattice_mps.Z2.add_charges(charges_x, charges_y)
        
        # initialize the variables to save
        errors_tr = [[0, 0]]
        errors = [0]
        entropies = [[0]]
        schmidt_vals = np.array([0]*chi)
        schmidt_vals[0] = 1
        schmidt_vals = [schmidt_vals]
        # ---------------------------------------------------------
        # Initial Observables
        # ---------------------------------------------------------
        if "wl" in args.obs:
            print(f"wilson loop before trotter - L:{L}, chi:{chi}")
            lattice_mps.Z2.wilson_Z2_dual(mpo_sites=args.sites, ls=args.ladders) #list(range(s))
            lattice_mps.w = lattice_mps.Z2.mpo.copy()
            if args.moment == 1:
                W.append(lattice_mps.mpo_first_moment().real)
            elif args.moment == 2:
                W.append(lattice_mps.mpo_second_moment().real)
            elif args.moment == 4:
                W.append(lattice_mps.mpo_fourth_moment().real)

        if  "el" in args.obs:
            print(f"electric field before trotter - L:{L}, chi:{chi}")
            E_h = np.zeros((2*args.l+1,2*L-1))
            E_h[:] = np.nan
            E_h = lattice_mps.electric_field_Z2(E_h)
            El.append(E_h)
        
        if "thooft" in args.obs:
            print(f"'t Hooft string before trotter - L:{L}, chi:{chi}")
            lattice_mps.Z2.thooft(site=args.sites, l=args.ladders, direction=direction)
            lattice_mps.w = lattice_mps.Z2.mpo.copy()
            S.append(lattice_mps.mpo_first_moment().real)

        if "mag" in args.obs:
            print(f"Magnetization before trotter - L:{L}, chi:{chi}")
            lattice_mps.order_param()
            if args.moment == 1:
                M.append(lattice_mps.mpo_first_moment().real/(len(lattice_mps.Z2.latt.plaquettes())-(2*(L-3)+2*(args.l))))
            elif args.moment == 2:
                M.append(lattice_mps.mpo_second_moment().real/(len(lattice_mps.Z2.latt.plaquettes())-(2*(L-3)+2*(args.l)))**2)
            elif args.moment == 4:
                M.append(lattice_mps.mpo_fourth_moment().real/(len(lattice_mps.Z2.latt.plaquettes())-(2*(L-3)+2*(args.l)))**4)

        if "corr" in args.obs:
            s = (charges_x[0]+charges_x[1]) //2
            lad = np.min(charges_y)
            print(f"connected correlator before trotter - L:{L}, chi:{chi}")
            corr = lattice_mps.connected_correlator(site=s, lad=lad)
            C.append(corr)
        
        if "eed" in args.obs:
            s = (charges_x[0]+charges_x[1]) //2
            print(f"electric energy density before trotter - L:{L}, chi:{chi}")
            ed = lattice_mps.electric_energy_density_Z2(site=s)
            Ed.append(ed)

        # ---------------------------------------------------------
        # Trotter Evolution
        # ---------------------------------------------------------        
        lattice_mps.enlarge_chi()
        lattice_mps.ancilla_sites = lattice_mps.sites.copy()
        for t in range(args.npoints):
            
            lattice_mps, error, entropy, schmidt_values = lattice_mps.TEBD_variational_Z2_trotter_step(trotter_step=t, delta=args.delta, h_ev=args.h_ev, where=args.where, bond=args.bond)
            
            # ---------------------------------------------------------
            # Trotter Observables
            # ---------------------------------------------------------  
            if "wl" in args.obs:
                print(f"wilson loop for trotter step:{t}, L:{L}, chi:{chi}")
                lattice_mps.Z2.wilson_Z2_dual(mpo_sites=args.sites, ls=args.ladders) #list(range(s))
                lattice_mps.w = lattice_mps.Z2.mpo.copy()
                if args.moment == 1:
                    W.append(lattice_mps.mpo_first_moment().real)
                elif args.moment == 2:
                    W.append(lattice_mps.mpo_second_moment().real)
                elif args.moment == 4:
                    W.append(lattice_mps.mpo_fourth_moment().real)

            if  "el" in args.obs:
                print(f"electric field for trotter step:{t}, L:{L}, chi:{chi}")
                E_h = np.zeros((2*args.l+1,2*L-1))
                E_h[:] = np.nan
                E_h = lattice_mps.electric_field_Z2(E_h)
                El.append(E_h)
            
            if "thooft" in args.obs:
                print(f"'t Hooft string for trotter step:{t}, L:{L}, chi:{chi}")
                lattice_mps.Z2.thooft(site=args.sites, l=args.ladders, direction=direction)
                lattice_mps.w = lattice_mps.Z2.mpo.copy()
                S.append(lattice_mps.mpo_first_moment().real)

            if "mag" in args.obs:
                print(f"Magnetization for trotter step:{t}, L:{L}, chi:{chi}")
                lattice_mps.order_param()
                if args.moment == 1:
                    M.append(lattice_mps.mpo_first_moment().real/(len(lattice_mps.Z2.latt.plaquettes())-(2*(L-3)+2*(args.l))))
                elif args.moment == 2:
                    M.append(lattice_mps.mpo_second_moment().real/(len(lattice_mps.Z2.latt.plaquettes())-(2*(L-3)+2*(args.l)))**2)
                elif args.moment == 4:
                    M.append(lattice_mps.mpo_fourth_moment().real/(len(lattice_mps.Z2.latt.plaquettes())-(2*(L-3)+2*(args.l)))**4)

            if "corr" in args.obs:
                print(f"connected correlator for trotter step:{t}, L:{L}, chi:{chi}")
                corr = lattice_mps.connected_correlator(site=s, lad=lad)
                C.append(corr)
            
            if "eed" in args.obs:
                s = (charges_x[0]+charges_x[1]) //2
                print(f"electric energy density before trotter - L:{L}, chi:{chi}")
                ed = lattice_mps.electric_energy_density_Z2(site=s)
                Ed.append(ed)

            print(f"\nError at trotter step: {t} is: {error}\n")
            errors_tr.append(error)
            errors.append(error[-1])
            entropies.append(entropy)
            schmidt_vals.append(schmidt_values)

        if "wl" in args.obs:
            np.save(
                        f"{parent_path}/results/wilson_loops/wilson_loop_{moment}_moment_{args.model}_direct_lattice_{args.l}x{L-1}_{sector}_{args.charges_x}-{args.charges_y}_h_i_{args.h_i}_h_ev_{args.h_ev}_delta_{args.delta}_trotter_steps_{args.npoints}_chi_{chi}.npy",
                        W,
                    )
        if "el" in args.obs:
            np.save(
                        f"{parent_path}/results/electric_field/electric_field_{args.model}_direct_lattice_{args.l}x{L-1}_{sector}_{args.charges_x}-{args.charges_y}_h_i_{args.h_i}_h_ev_{args.h_ev}_delta_{args.delta}_trotter_steps_{args.npoints}_chi_{chi}.npy",
                        El,
                    )
        if "thooft" in args.obs:
            np.save(
                        f"{parent_path}/results/thooft/thooft_string_{moment}_moment_{args.sites[0]}-{args.ladders[0]}_{direction}_{args.model}_direct_lattice_{args.l}x{L-1}_{sector}_{args.charges_x}-{args.charges_y}_h_i_{args.h_i}_h_ev_{args.h_ev}_delta_{args.delta}_trotter_steps_{args.npoints}_chi_{chi}.npy",
                        S,
                    )
        if "mag" in args.obs:
            np.save(
                        f"{parent_path}/results/mag_data/dual_mag_{moment}_moment_{args.model}_direct_lattice_{args.l}x{L-1}_{sector}_{args.charges_x}-{args.charges_y}_h_i_{args.h_i}_h_ev_{args.h_ev}_delta_{args.delta}_trotter_steps_{args.npoints}_chi_{chi}.npy",
                        M,
                    )
        if "corr" in args.obs:
            np.save(
                        f"{parent_path}/results/mag_data/connected_correlator_s_{s}_l_{lad}_{args.model}_direct_lattice_{args.l}x{L-1}_{sector}_{args.charges_x}-{args.charges_y}_h_i_{args.h_i}_h_ev_{args.h_ev}_delta_{args.delta}_trotter_steps_{args.npoints}_chi_{chi}.npy",
                        C,
                    )
        if "eed" in args.obs:
            np.save(
                        f"{parent_path}/results/electric_field/electric_energy_density_s_{s}_{args.model}_direct_lattice_{args.l}x{L-1}_{sector}_{args.charges_x}-{args.charges_y}_h_i_{args.h_i}_h_ev_{args.h_ev}_delta_{args.delta}_trotter_steps_{args.npoints}_chi_{chi}.npy",
                        Ed,
                    )
            
        if args.bond == False:
            args.where = "all"

        if args.training:
            save_list_of_lists(
                f"{parent_path}/results/error_data/errors_tr_quench_dynamics_{args.model}_direct_lattice_{args.l}x{L-1}_{sector}_{charges_x}-{charges_y}_h_i_{args.h_i}_h_ev_{args.h_ev}_delta_{args.delta}_trotter_steps_{args.npoints}_chi_{chi}",
                errors_tr,
            )
            np.savetxt(
                f"{parent_path}/results/error_data/errors_quench_dynamics_{args.model}_direct_lattice_{args.l}x{L-1}_{sector}_{charges_x}-{charges_y}_h_i_{args.h_i}_h_ev_{args.h_ev}_delta_{args.delta}_trotter_steps_{args.npoints}_chi_{chi}",
                errors,
            )
        else:
            np.savetxt(
                f"{parent_path}/results/error_data/errors_quench_dynamics_{args.model}_direct_lattice_{args.l}x{L-1}_{sector}_{charges_x}-{charges_y}_h_i_{args.h_i}_h_ev_{args.h_ev}_delta_{args.delta}_trotter_steps_{args.npoints}_chi_{chi}",
                errors,
            )
            
        save_list_of_lists(
            f"{parent_path}/results/entropy_data/{args.where}_bond_entropy_quench_dynamics_{args.model}_direct_lattice_{args.l}x{L-1}_{sector}_{charges_x}-{charges_y}_h_i_{args.h_i}_h_ev_{args.h_ev}_delta_{args.delta}_trotter_steps_{args.npoints}_chi_{chi}",
            entropies,
        )
        save_list_of_lists(
            f"{parent_path}/results/entropy_data/{args.where}_schmidt_vals_quench_dynamics_{args.model}_direct_lattice_{args.l}x{L-1}_{sector}_{charges_x}-{charges_y}_h_i_{args.h_i}_h_ev_{args.h_ev}_delta_{args.delta}_trotter_steps_{args.npoints}_chi_{chi}",
            schmidt_vals,
        )
        if args.where == "all":
            entropy_mid = access_txt(
                f"{parent_path}/results/entropy_data/{args.where}_bond_entropy_quench_dynamics_{args.model}_direct_lattice_{args.l}x{L-1}_{sector}_{charges_x}-{charges_y}_h_i_{args.h_i}_h_ev_{args.h_ev}_delta_{args.delta}_trotter_steps_{args.npoints}_chi_{chi}",
                (L-1) // 2,
            )
            np.savetxt(
                f"{parent_path}/results/entropy_data/{L // 2}_bond_entropy_quench_dynamics_{args.model}_direct_lattice_{args.l}x{L-1}_{sector}_{charges_x}-{charges_y}_h_i_{args.h_i}_h_ev_{args.h_ev}_delta_{args.delta}_trotter_steps_{args.npoints}_chi_{chi}",
                entropy_mid,
            )



import numpy as np
import argparse
from qs_mps.sparse_hamiltonians_and_operators import exact_evolution_sparse
from qs_mps.utils import get_precision, save_list_of_lists, access_txt
from qs_mps.applications.Z2.ground_state_multiprocessing import ground_state_Z2_exact, ground_state_Z2_exact_test
from qs_mps.applications.Z2.exact_hamiltonian import H_Z2_gauss

# EXACT DIAGONALIZATION to find ground states of the Z2 Pure Gauge Theory 
# changing the plaquette (magnetic) parameters in its direct formulation

parser = argparse.ArgumentParser(prog="gs_search_Z2_exact")
parser.add_argument("l", help="Number of ladders in the direct lattice", type=int)
parser.add_argument("L", help="Number of rungs per ladder", type=int)
parser.add_argument(
    "npoints",
    help="Number of points in an interval of transverse field values",
    type=int,
)
parser.add_argument(
    "h_i", help="Starting value of h (external transverse field on the dual lattice)", type=float
)
parser.add_argument(
    "h_f", help="Final value of h (external transverse field on the dual lattice)", type=float
)
parser.add_argument(
    "path",
    help="Path to the drive depending on the device used. Available are 'pc', 'mac', 'marcos'",
    type=str,
)
parser.add_argument("-cx", "--charges_x", help="a list of the first index of the charges", nargs="*", type=int)
parser.add_argument("-cy", "--charges_y", help="a list of the second index of the charges", nargs="*", type=int)
parser.add_argument(
    "-m", "--model", help="Model to simulate. By default Z2_dual", default="Z2_dual", type=str
)
parser.add_argument(
    "-v",
    "--save",
    help="Save the ground states. By default True",
    action="store_false",
)
parser.add_argument(
    "-p",
    "--sparse",
    help="Use Sparse method for exact diagonalization. By default True",
    action="store_false",
)
parser.add_argument(
    "-w",
    "--spectrum",
    help="The spectrum of the Exact Hamiltonian. By default (-1) computes only the ground state eigenvalue. Can compute the spectrum up to the degrees of freedom of the dual lattice",
    default=-1,
    type=int,
)
parser.add_argument(
    "-U", "--gauss", help="Gauss constraint parameter. By default 1e+3", default=1e+3, type=float
)

args = parser.parse_args()

# define the interval
interval = np.linspace(args.h_i, args.h_f, args.npoints)

# take the path and precision to save files
# if we want to save the tensors we save them locally because they occupy a lot of memory
if args.path == "pc":
    parent_path = "G:/My Drive/projects/1_Z2"
    path_eigvec = "D:/code/projects/1_Z2"
elif args.path == "mac":
    parent_path = "/Users/fradm98/Google Drive/My Drive/projects/1_Z2"
    path_eigvec = "/Users/fradm98/Desktop/projects/1_Z2"
elif args.path == "marcos":
    parent_path = "/Users/fradm/Google Drive/My Drive/projects/1_Z2"
    path_eigvec = "/Users/fradm/Desktop/projects/1_Z2"
else:
    raise SyntaxError("Path not valid. Choose among 'pc', 'mac', 'marcos'")

# define the precision
num = (args.h_f - args.h_i) / args.npoints
precision = get_precision(num)

# define the initial guess state
dof_direct = (2*args.l*args.L - args.l - args.L)
# constant without sparse method (if we start from h=0)
c = (1/np.sqrt(2))**dof_direct

print("finding guess for the sparse computation...")
v0 = np.full(2**dof_direct, c)
v0 = np.array(v0, dtype=complex)
print(v0)

# choose how many eigenvalues to compute
if args.spectrum == -1:
    spectrum = "gs"
elif args.spectrum == 0:
    spectrum = "all"

    
# define the sector by looking of the given charges
if len(args.charges_x) == 0:
    sector = "vacuum_sector"
    args.charges_x = None
    args.charges_y = None
else:
    for i in range(1,args.l*args.L):
        if len(args.charges_x) == i:
            sector = f"{i}_particle(s)_sector"

# ---------------------------------------------------------
# Exact Diagonalization
# ---------------------------------------------------------
args_lattice = {
    "L": args.L,
    "l": args.l,
    "model": args.model,
    "path": path_eigvec,
    "save": args.save,
    # "v0": v0,
    "sparse": args.sparse,
    "precision": precision,
    "spectrum": spectrum,
    "U": args.gauss,
    "sector": sector,
    "charges_x": args.charges_x,
    "charges_y": args.charges_y,
}

energy = []

for h in interval:
    print(f"computing ground state for param: {h:.{precision}f}")
    Z2 = H_Z2_gauss(l=args.l, L=args.L, model=args.model, lamb=h, U=args.gauss)
    if len(args.charges_x) > 0:
        Z2.add_charges(rows=args.charges_x, columns=args.charges_y)
        Z2._define_sector()
    e, v = Z2.diagonalize(v0=v0, sparse=args.sparse, save=args.save, path=path_eigvec, cx=args.charges_x, cy=args.charges_y, precision=precision) # , path=args.path, precision=precision, spectrum=spectrum, cx=args.charges_x, cy=args.charges_y
    energy.append(e[0])
    v0 = v[:,0]
    print("groud state:")
    print(v0)

# energy = ground_state_Z2_exact_test(
#     args_lattice=args_lattice, param=interval
# )

if spectrum == "all":
    save_list_of_lists(
        f"{parent_path}/results/exact/energy_data/energies_{args.model}_direct_lattice_{args.l-1}x{args.L-1}_{sector}_{args.charges_x}-{args.charges_y}_h_{args.h_i}-{args.h_f}_delta_{args.npoints}",
        energy,
    )
    energy_gs = access_txt(
            f"{parent_path}/results/exact/energy_data/energies_{args.model}_direct_lattice_{args.l-1}x{args.L-1}_{sector}_{args.charges_x}-{args.charges_y}_h_{args.h_i}-{args.h_f}_delta_{args.npoints}",
            0,
        )
    np.savetxt(f"{parent_path}/results/exact/energy_data/energies_{args.model}_direct_lattice_{args.l-1}x{args.L-1}_{sector}_{args.charges_x}-{args.charges_y}_h_{args.h_i}-{args.h_f}_delta_{args.npoints}",
        energy_gs,
    )
else:
    np.savetxt(f"{parent_path}/results/exact/energy_data/energies_{args.model}_direct_lattice_{args.l-1}x{args.L-1}_{sector}_{args.charges_x}-{args.charges_y}_h_{args.h_i}-{args.h_f}_delta_{args.npoints}",
        energy,
    )


import numpy as np

import scipy.linalg as la
import scipy.sparse.linalg as spla

from .utils import get_middle_chain_schmidt_values, von_neumann_entropy, mps_to_vector
from scipy.sparse import identity, csc_array, csc_matrix

# -----------------------------------------------
# Sparse Non-Diagonal Pauli Indices
# -----------------------------------------------
def sparse_non_diag_paulis_indices(n: int, N: int):
    """
    Returns a tuple (row_indices, col_indices) containing the row and col indices of the non_zero elements
    of the tensor product of a non diagonal pauli matrix (x, y) acting over a single qubit in a Hilbert
    space of N qubits

    """
    if 0 <= n < N:
        block_length = 2 ** (N - n - 1)
        nblocks = 2**n
        ndiag_elements = block_length * nblocks
        k = np.arange(ndiag_elements, dtype=int)
        red_row_col_ind = (k % block_length) + 2 * (k // block_length) * block_length
        upper_diag_row_indices = red_row_col_ind
        upper_diag_col_indices = block_length + red_row_col_ind
        row_indices = np.concatenate((upper_diag_row_indices, upper_diag_col_indices))
        col_indices = np.concatenate((upper_diag_col_indices, upper_diag_row_indices))
        return row_indices, col_indices
    else:
        raise ValueError("Index n must fulfill 0 <= n < N")

# ---------------------------------------------------------------------------------------
# Sparse Pauli X
# ---------------------------------------------------------------------------------------
def sparse_pauli_x(n: int, L: int, row_indices_cache: np.ndarray=None, col_indices_cache: np.ndarray=None):
    """
    Returns a CSC sparse matrix representation of the pauli_x matrix acting over qubit n in a Hilbert space of L qubits
    0 <= n < L

    """
    if 0 <= n < L:
        if (row_indices_cache is None) or (col_indices_cache is None):
            row_indices_cache, col_indices_cache = sparse_non_diag_paulis_indices(n, L)
        data = np.ones_like(row_indices_cache)
        result = csc_array(
            (data, (row_indices_cache, col_indices_cache)), shape=(2**L, 2**L)
        )  # , dtype=complex
        return result
    else:
        raise ValueError("Index n must fulfill 0 <= n < L")

# ---------------------------------------------------------------------------------------
# Sparse Pauli Y
# ---------------------------------------------------------------------------------------
def sparse_pauli_y(n: int, L: int, row_indices_cache: np.ndarray=None, col_indices_cache: np.ndarray=None):
    """
    Returns a CSC sparse matrix representation of the pauli_y matrix acting over qubit n in a Hilbert space of L qubits
    0 <= n < L

    """
    if 0 <= n < L:
        if (row_indices_cache is None) or (col_indices_cache is None):
            row_indices_cache, col_indices_cache = sparse_non_diag_paulis_indices(n, L)
        data = -1j * np.ones_like(row_indices_cache)
        data[len(data) // 2 : :] = 1j
        result = csc_array(
            (data, (row_indices_cache, col_indices_cache)),
            shape=(2**L, 2**L),
            dtype=complex,
        )
        return result
    else:
        raise ValueError("Index n must fulfill 0 <= n < L")


# ---------------------------------------------------------------------------------------
# Sparse Pauli Z
# ---------------------------------------------------------------------------------------
def sparse_pauli_z(n: int, L: int):
    """
    Returns a CSC sparse matrix representation of the pauli_z matrix acting over qubit n in a Hilbert space of L qubits
    0 <= n < L

    """
    if 0 <= n < L:
        block_length = 2 ** (L - n)
        nblocks = 2**n
        block = np.ones(block_length, dtype=int)
        block[block_length // 2 : :] = -1
        diag = np.tile(block, nblocks)
        row_col_indices = np.arange(2**L, dtype=int)
        result = csc_array(
            (diag, (row_col_indices, row_col_indices)),
            shape=(2**L, 2**L),
            dtype=complex,
        )
        return result
    else:
        raise ValueError("Index n must fulfill 0 <= n < L")

# ---------------------------------------------------------------------------------------
# Sparse Magnetization
# ---------------------------------------------------------------------------------------
def sparse_magnetization(L, op="X", staggered: bool=False):
    if op == "X":
        op = sparse_pauli_x
    elif op == "Z":
        op = sparse_pauli_z

    m = 0
    c = [1 for _ in range(L)]
    if staggered:
        c = [(-1)**(i//2) for i in range(L)]
    for i in range(L):
        n_row_indices, n_col_indices = sparse_non_diag_paulis_indices(i, L)
        m += c[i] * op(i, L, n_row_indices, n_col_indices)
    return m/L

# -----------------------------------------------
# Sparse Ising Hamiltonian
# -----------------------------------------------
def sparse_ising_hamiltonian(J: float, h_t: float, h_l: float, L: int, long: str="X"):
    """
    Returns a sparse representation of the Hamiltonian of the 1D Ising model in a chain of length L
    with open boundary conditions
    J < 0: Antiferromagnetic case (Unique ground state of total angular momentum S=0)
    J > 0: Ferromagnetic case (2-fold degeneracy of the ground state of angular momentum S=L/2)
    """
    hamiltonian_l = csc_array((2**L, 2**L), dtype=complex)
    hamiltonian_t = csc_array((2**L, 2**L), dtype=complex)
    hamiltonian_int = csc_array((2**L, 2**L), dtype=complex)
    
    if long == "X":
    # First sum over the terms containing sigma_x, sigma_y because the non-zero element indices are the same
    # so that this improves performance
        if h_t != 0:
            for n in range(L):
                n_pauli_z = sparse_pauli_z(n, L)
                hamiltonian_t += n_pauli_z
        if h_l != 0:
            for n in range(L):
                n_row_indices, n_col_indices = sparse_non_diag_paulis_indices(n, L)
                n_pauli_x = sparse_pauli_x(n, L, n_row_indices, n_col_indices)
                hamiltonian_l += n_pauli_x

        # Sum over sigma_z terms
        for n in range(L - 1):
            n_row_indices, n_col_indices = sparse_non_diag_paulis_indices(n, L)
            np1_row_indices, np1_col_indices = sparse_non_diag_paulis_indices(n+1, L)
            n_pauli_x = sparse_pauli_x(n, L, n_row_indices, n_col_indices)
            np1_pauli_x = sparse_pauli_x(n + 1, L, np1_row_indices, np1_col_indices)
            hamiltonian_int += n_pauli_x @ np1_pauli_x

    if long == "Z":    
    # First sum over the terms containing sigma_x, sigma_y because the non-zero element indices are the same
    # so that this improves performance
        if h_t != 0:
            for n in range(L):
                n_row_indices, n_col_indices = sparse_non_diag_paulis_indices(n, L)
                n_pauli_x = sparse_pauli_x(n, L, n_row_indices, n_col_indices)
                hamiltonian_t += n_pauli_x
        if h_l != 0:
            for n in range(L):
                n_pauli_z = sparse_pauli_z(n, L)
                hamiltonian_l += n_pauli_z

        # Interaction
        for n in range(L - 1):
            n_pauli_z = sparse_pauli_z(n, L)
            np1_pauli_z = sparse_pauli_z(n + 1, L)
            hamiltonian_int += n_pauli_z @ np1_pauli_z

    return -J * hamiltonian_int - h_t * hamiltonian_t - h_l * hamiltonian_l

# -----------------------------------------------
# Sparse ANNNI Hamiltonian
# -----------------------------------------------
def sparse_ANNNI_hamiltonian(J: float, h_t: float, h_ll: float, L: int, eps: float=1e-5, long: str="X", deg_method: int=1):
    """
    sparse_ANNNI_hamiltonian

    This function gives a representation of the 1D Axial Next Nearest Neighbor Interaction model.
    The next nearest neighbor interaction (h_ll) is competing with the nearest neighbor interaction (J)
    We use eps to break the degeneracy for small transverse field h_t.

    """
    hamiltonian_ll = csc_array((2**L, 2**L), dtype=complex)
    hamiltonian_deg = csc_array((2**L, 2**L), dtype=complex)
    hamiltonian_t = csc_array((2**L, 2**L), dtype=complex)
    hamiltonian_int = csc_array((2**L, 2**L), dtype=complex)
    if long == "X":
        if h_t != 0:
            # transverse field
            for n in range(L):
                n_pauli_z = sparse_pauli_z(n, L)
                hamiltonian_t += n_pauli_z
        if h_ll != 0:
            # next nearest neighbor interaction
            for n in range(L-2):
                n_row_indices, n_col_indices = sparse_non_diag_paulis_indices(n, L)
                np2_row_indices, np2_col_indices = sparse_non_diag_paulis_indices(n+2, L)
                n_pauli_x = sparse_pauli_x(n, L, n_row_indices, n_col_indices)
                np2_pauli_x = sparse_pauli_x(n+2, L, np2_row_indices, np2_col_indices)
                hamiltonian_ll += n_pauli_x @ np2_pauli_x
 
        # nearest neighbor interaction
        for n in range(L-1):
            n_row_indices, n_col_indices = sparse_non_diag_paulis_indices(n, L)
            np1_row_indices, np1_col_indices = sparse_non_diag_paulis_indices(n+1, L)
            n_pauli_x = sparse_pauli_x(n, L, n_row_indices, n_col_indices)
            np1_pauli_x = sparse_pauli_x(n+1, L, np1_row_indices, np1_col_indices)
            hamiltonian_int += n_pauli_x @ np1_pauli_x
        if eps != 0:
            # add a term to break the double degeneracy of the ground state
            if deg_method == 0:
                hamiltonian_deg = 0
            elif deg_method == 1:
                hamiltonian_deg = sparse_pauli_x(n=0,L=L) - identity(2**L)
            elif deg_method == 2:
                for n in range(L):
                    hamiltonian_deg += (1 + (-1)**(n//2)) * sparse_pauli_x(n=n,L=L) -  2 * identity(2**L)
            elif deg_method == 3:
                for n in range(L):
                    hamiltonian_deg += (-1)**(n//2) * sparse_pauli_x(n=n,L=L) - identity(2**L)
            else:
                raise ValueError("Choose a proper degeneracy method")
    return - J * hamiltonian_int + h_ll * hamiltonian_ll - h_t * hamiltonian_t - eps * hamiltonian_deg


# -----------------------------------------------
# Sparse Cluster Hamiltonian
# -----------------------------------------------
def sparse_cluster_hamiltonian(J: float, h_t: float, L: int, eps: float=1e-5, long: str="X"):
    """
    Returns a sparse representation of the Hamiltonian of the 1D Cluster model in a chain of length L
    with open boundary conditions
    J < 0: Antiferromagnetic case (Unique ground state of total angular momentum S=0)
    J > 0: Ferromagnetic case (L+1-fold degeneracy of the ground state of angular momentum L/2) -> Dicke states for even L
    """
    hamiltonian_t = csc_array((2**L, 2**L), dtype=complex)
    hamiltonian_deg = csc_array((2**L, 2**L), dtype=complex)
    hamiltonian_int = csc_array((2**L, 2**L), dtype=complex)
    
    if long == "X":
    # First sum over the terms containing sigma_x, sigma_y because the non-zero element indices are the same
    # so that this improves performance
        if h_t != 0:
            for n in range(L):
                n_pauli_z = sparse_pauli_z(n, L)
                hamiltonian_t += n_pauli_z
        if eps != 0:
            for n in range(L):
                n_row_indices, n_col_indices = sparse_non_diag_paulis_indices(n, L)
                n_pauli_x = sparse_pauli_x(n, L, n_row_indices, n_col_indices)
                hamiltonian_deg += n_pauli_x
            
        # Interaction
        for n in range(L - 2):
            n_row_indices, n_col_indices = sparse_non_diag_paulis_indices(n, L)
            np2_row_indices, np2_col_indices = sparse_non_diag_paulis_indices(n+2, L)
            n_pauli_x = sparse_pauli_x(n, L, n_row_indices, n_col_indices)
            np1_pauli_z = sparse_pauli_z(n + 1, L)
            np2_pauli_x = sparse_pauli_x(n + 2, L, np2_row_indices, np2_col_indices)
            hamiltonian_int += n_pauli_x @ np1_pauli_z @ np2_pauli_x

    return -J * hamiltonian_int - h_t * hamiltonian_t - eps * hamiltonian_deg

# -----------------------------------------------
# Sparse Cluster-XY Hamiltonian
# -----------------------------------------------
def sparse_cluster_xy_hamiltonian(J: float, h_t: float, h_x: float, h_y: float, L: int, eps: float=1e-5, long: str="X"):
    """
    Returns a sparse representation of the Hamiltonian of the 1D Cluster-XY model in a chain of length L
    with open boundary conditions
    J < 0: Antiferromagnetic case (Unique ground state of total angular momentum S=0)
    J > 0: Ferromagnetic case (L+1-fold degeneracy of the ground state of angular momentum L/2) -> Dicke states for even L
    """
    hamiltonian_t = csc_array((2**L, 2**L), dtype=complex)
    hamiltonian_x = csc_array((2**L, 2**L), dtype=complex)
    hamiltonian_y = csc_array((2**L, 2**L), dtype=complex)
    hamiltonian_deg = csc_array((2**L, 2**L), dtype=complex)
    hamiltonian_int = csc_array((2**L, 2**L), dtype=complex)
    
    if long == "X":
    # First sum over the terms containing sigma_x, sigma_y because the non-zero element indices are the same
    # so that this improves performance
        if h_t != 0:
            for n in range(L):
                n_pauli_z = sparse_pauli_z(n, L)
                hamiltonian_t += n_pauli_z
        if h_x != 0:
            for n in range(L-1):
                n_row_indices, n_col_indices = sparse_non_diag_paulis_indices(n, L)
                n_pauli_x = sparse_pauli_x(n, L, n_row_indices, n_col_indices)
                hamiltonian_x += n_pauli_x
        if h_y != 0:
            for n in range(L-1):
                n_row_indices, n_col_indices = sparse_non_diag_paulis_indices(n, L)
                n_pauli_y = sparse_pauli_y(n, L, n_row_indices, n_col_indices)
                hamiltonian_y += n_pauli_y
        if eps != 0:
            for n in range(L):
                n_row_indices, n_col_indices = sparse_non_diag_paulis_indices(n, L)
                n_pauli_x = sparse_pauli_x(n, L, n_row_indices, n_col_indices)
                hamiltonian_deg += n_pauli_x
            
        # Interaction
        for n in range(L - 2):
            n_row_indices, n_col_indices = sparse_non_diag_paulis_indices(n, L)
            np2_row_indices, np2_col_indices = sparse_non_diag_paulis_indices(n+2, L)
            n_pauli_x = sparse_pauli_x(n, L, n_row_indices, n_col_indices)
            np1_pauli_z = sparse_pauli_z(n + 1, L)
            np2_pauli_x = sparse_pauli_x(n + 2, L, np2_row_indices, np2_col_indices)
            hamiltonian_int += n_pauli_x @ np1_pauli_z @ np2_pauli_x

    return -J * hamiltonian_int - h_t * hamiltonian_t - eps * hamiltonian_deg + h_x * hamiltonian_x + h_y * hamiltonian_y

# ---------------------------------------------------------------------------------------
# Diagonalization
# ---------------------------------------------------------------------------------------
def diagonalization(H: csc_matrix, sparse: bool, v0: np.ndarray=None, k: int=1, which: str='SA'):
    if sparse:
        e,v = spla.eigsh(H, k=k, which=which, v0=v0)
    else:
        e,v = la.eigh(H.toarray())
    return e,v

# ---------------------------------------------------------------------------------------
# Sparse Ground state
# ---------------------------------------------------------------------------------------
def sparse_ising_ground_state(
    L: int, h_t: float, h_l: float = 1e-7, J: float = 1, k: int = 1
) -> csc_array:
    """
    exact_initial_state

    This function is computing the initial state given by an Ising Hamiltonian.

    L: int - chain size
    h_t: float - initial transverse field parameter
    h_l: float - initial longitudinal field parameter
    k: int - number of eigenvalues we want to compute. By default 1

    """
    print("Finding the Hamiltonian...")
    H = sparse_ising_hamiltonian(J=J, h_t=h_t, h_l=h_l, L=L)
    print("Hamiltonian found")
    e, v = diagonalization(H, sparse=True, k=k)
    print(f"first {k} eigenvalue(s) SA (Smallest (algebraic) eigenvalues): {e}")
    psi = v[:, 0]
    return psi

# ---------------------------------------------------------------------------------------
# Sparse U Evolution
# ---------------------------------------------------------------------------------------
def U_evolution_sparse(
    psi_init: csc_array,
    H_ev: csc_array,
    trotter: int,
    time: float,
):
    """
    U_evolution

    This function applies a time evolution operator to some initial state.
    The evolution operator uses the Ising hamiltonian with some tunable parameters.

    psi_init: csc_array - initial state to be evolved
    H_ev: csc_array - Time evolution Hamiltonian
    trotter: int - the trotter step we are during the evolution
    time: float - indicates the final time we want to reach

    """
    delta = time / trotter
    H_ev = -1j * delta * H_ev
    psi_ev = spla.expm_multiply(H_ev, psi_init)
    return psi_ev

# ---------------------------------------------------------------------------------------
# Sparse exact Evolution
# ---------------------------------------------------------------------------------------
def exact_evolution_sparse(
    L: int,
    h_t: float,
    h_ev: float,
    time: float,
    trotter_steps: int,
    h_l: float = 1e-7,
    flip: bool = False,
    where: int = -1,
    bond: bool = True,
    model: str = None,
):
    """
    exact_evolution

    This function evolves an initial state for trotter steps times.
    We extract at each time step the local and total magnetization.

    L: int - chain size
    trotter_step: int - indicates the specific trotter step we are evolving
    time: float - indicates the final time we want to reach
    h_t: float - initial transverse field parameter
    h_l: float - initial longitudinal field parameter
    h_ev: float - evolution transverse field parameter
    where: int - in which bond we perform the Schmidt decompositon
    bond: bool - if we compute the entropy for all the bonds. By default False

    """
    if bond:
        entropy = [0]
    else:
        entropy = [0] * (L - 3)

    # psi_exact = sparse_ising_ground_state(L=L, h_t=h_t, h_l=h_l, k=6).reshape(2**L, 1)
    # for now only all spin up
    # psi_exact = np.zeros((2**L, 1))
    # psi_exact[0] = 1
    init_state = np.zeros((1, 2, 1))
    init_state[0, 0, 0] = 1
    mps = [init_state for _ in range(L)]
    psi_exact = mps_to_vector(mps).reshape(2**L, 1)
    if flip:
        flip = sparse_pauli_x(n=L // 2, L=L)
        psi_exact = csc_array(flip @ psi_exact)
    # local Z
    mag_loc_Z_op = [sparse_pauli_z(n=i, L=L) for i in range(L)]
    # local X
    mag_loc_X_op = sparse_pauli_x(n=L // 2, L=L)
    # total
    mag_tot_op = sparse_ising_hamiltonian(L=L, h_t=0, h_l=-1, J=0)

    mag_exact_loc_Z = []
    mag_exact_loc_X = []
    mag_exact_tot = []

    # local Z
    mag_exact = []
    for i in range(L):
        mag_exact.append(
            (psi_exact.T.conjugate() @ mag_loc_Z_op[i] @ psi_exact)[0, 0].real
        )
    mag_exact_loc_Z.append(mag_exact)

    # local X
    mag_exact_loc_X.append(
        (psi_exact.T.conjugate() @ mag_loc_X_op @ psi_exact)[0, 0].real
    )

    # total
    mag_exact_tot.append((psi_exact.T.conjugate() @ mag_tot_op @ psi_exact)[0, 0].real)

    H_ev = sparse_ising_hamiltonian(L=L, J=1, h_l=h_l, h_t=h_ev)

    entropy_tot = []
    entropy_tot.append(entropy)

    psi_new = psi_exact
    for trott in range(trotter_steps):
        print(f"-------- Trotter step {trott} ---------")
        # exact
        psi_new = U_evolution_sparse(
            psi_init=psi_new, H_ev=H_ev, trotter=trotter_steps, time=time
        )
        # entropy
        sing_vals = get_middle_chain_schmidt_values(psi_new, where=where, bond=bond)
        entropy = []
        for s in sing_vals:
            ent = von_neumann_entropy(s)
            entropy.append(ent)
        entropy_tot.append(entropy)
        # local Z
        mag_exact = []
        for i in range(L):
            mag_exact.append(
                (psi_new.T.conjugate() @ mag_loc_Z_op[i] @ psi_new)[0, 0].real
            )
        mag_exact_loc_Z.append(mag_exact)

        # local X
        mag_exact_loc_X.append(
            (psi_new.T.conjugate() @ mag_loc_X_op @ psi_new)[0, 0].real
        )

        # total
        mag_exact_tot.append((psi_new.T.conjugate() @ mag_tot_op @ psi_new)[0, 0].real)
    return psi_new, mag_exact_loc_Z, mag_exact_loc_X, mag_exact_tot, entropy_tot


import concurrent.futures
import os
from qs_mps.mps_class import MPS
from qs_mps.applications.Z2.exact_hamiltonian import H_Z2_gauss
from qs_mps.utils import tensor_shapes
import signal
import numpy as np

# Define a function to handle the timeout
def timeout_handler(signum, frame):
    raise TimeoutError("Algorithm took too long to execute")

# Set the signal handler
signal.signal(signal.SIGALRM, timeout_handler)

def ground_state_Z2_exact_param(params):
    args_lattice = params[0]
    param = params[1]
    Z2 = H_Z2_gauss(L=args_lattice["L"], l=args_lattice["l"], model=args_lattice["model"], U=args_lattice["U"], lamb=param)
    if len(args_lattice["charges_x"]) > 0:
        # print(Z2.charges)
        Z2.add_charges(rows=args_lattice["charges_x"], columns=args_lattice["charges_y"])
        # print(Z2.charges)
        Z2._define_sector()
        print("before diag")
    e, v = Z2.diagonalize(v0=args_lattice["v0"], path=args_lattice["path"], save=args_lattice["save"], precision=args_lattice["precision"], cx=args_lattice["charges_x"], cy=args_lattice["charges_y"], sparse=args_lattice["sparse"])
    return e, v

def ground_state_Z2_exact(args_lattice, param):
    energies_param = []
    prec = args_lattice["precision"]
    for p in param:
        print(f"computing ground state for param: {p:.{prec}f}")
        params = [args_lattice, p]
        print(args_lattice["v0"])
        energy, vectors = ground_state_Z2_exact_param(params=params)
        energies_param.append(energy)
        v0 = vectors[:,0]
        args_lattice["v0"] = v0

    return energies_param

def ground_state_Z2_exact_test(args_lattice, param):
    energies_param = []
    prec = args_lattice["precision"]
    for p in param:
        print(f"computing ground state for param: {p:.{prec}f}")
        Z2 = H_Z2_gauss(L=args_lattice["L"], l=args_lattice["l"], model=args_lattice["model"], U=args_lattice["U"], lamb=p)
        if len(args_lattice["charges_x"]) > 0:
            print("adding charges")
            Z2.add_charges(rows=args_lattice["charges_x"], columns=args_lattice["charges_y"])
            Z2._define_sector()       
        energy, vectors = Z2.diagonalize(v0=args_lattice["v0"], sparse=args_lattice["sparse"], path=args_lattice["path"], save=args_lattice["save"], precision=args_lattice["precision"], cx=args_lattice["charges_x"], cy=args_lattice["charges_y"])
        energies_param.append(energy)
        v0 = vectors[:,0]
        args_lattice["v0"] = v0
        print(args_lattice["v0"])

    return energies_param

def ground_state_Z2_param(params):
    args_mps = params[0]
    param = params[1]
    ladder = MPS(
        L=args_mps["L"],
        d=args_mps["d"],
        model=args_mps["model"],
        chi=args_mps["chi"],
        h=param,
    )
    save = args_mps["save"]
    precision = args_mps["precision"]
    if ladder.model == "Z2_dual":
        ladder.L = ladder.L - 1
        if args_mps["sector"] != "vacuum_sector":
            ladder.Z2.add_charges(rows=args_mps["charges_x"], columns=args_mps["charges_y"])
            print(ladder.Z2.charges)
    if args_mps["guess"] == []:
        ladder._random_state(seed=3, chi=args_mps["chi"], type_shape=args_mps["type_shape"])
        tensor_shapes(ladder.sites)
        ladder.canonical_form(trunc_chi=True, trunc_tol=False)
    else:
        ladder.sites = args_mps["guess"].copy()
        ladder.enlarge_chi()

    energy, entropy, schmidt_vals, t_dmrg = ladder.DMRG(
        trunc_tol=args_mps["trunc_tol"],
        trunc_chi=args_mps["trunc_chi"],
        where=args_mps["where"],
        bond=args_mps["bond"],
        n_sweeps=args_mps["n_sweeps"],
        conv_tol=args_mps["conv_tol"]
    )

    print(f"energy of h:{param:.{precision}f}, L:{ladder.L} is:\n {energy}")
    print(f"Schmidt values in the middle of the chain:\n {schmidt_vals}")
    
    if not args_mps["training"]:
        energy = energy[-1]

    if save:
        ladder.save_sites(args_mps["path"], args_mps["precision"], args_mps["charges_x"], args_mps["charges_y"])
    args_mps["guess"] = ladder.sites.copy()
    return energy, entropy, schmidt_vals, t_dmrg


def ground_state_Z2_multpr(args_mps, multpr_param, cpu_percentage=90):
    max_workers = int(os.cpu_count() * (cpu_percentage / 100))
    with concurrent.futures.ProcessPoolExecutor(max_workers=max_workers) as executor:
        args = [[args_mps, param] for param in multpr_param]
        results = executor.map(ground_state_Z2_param, args)

    energies = []
    entropies = []
    i = 0
    for result in results:
        print(f"energy of h:{multpr_param[i]} is:\n {result[0][-1]}")
        energies.append(result[0])
        print(f"entropies of h:{multpr_param[i]} is:\n {result[1]}")
        entropies.append(result[1])
        i += 1
    return energies, entropies


def ground_state_Z2(args_mps, multpr, param):
    if multpr:
        energies_param, entropies_param = ground_state_Z2_multpr(
            args_mps=args_mps, multpr_param=param
        )
    else:
        energies_param = []
        entropies_param = []
        schmidt_vals_param = []
        time_param = []
        new_timeout_secs = 5
        for p in param:
            params = [args_mps, p]
            # Set the timeout period (in seconds)
            timeout_secs = new_timeout_secs # You can change this value according to your requirement
            # Set the alarm
            signal.alarm(int(timeout_secs+1))
            print(f"New timeout seconds: {int(timeout_secs+1)}")
            try:
                print("Running with guess state:")
                energy, entropy, schmidt_vals, t_dmrg = ground_state_Z2_param(params=params)

            except TimeoutError as e:
                print(e)
                args_mps["guess"] = []
                print("Running with random state:")
                energy, entropy, schmidt_vals, t_dmrg = ground_state_Z2_param(params=params)
            else:
                # Cancel the alarm if the algorithm finishes before the timeout
                signal.alarm(0)

            time_param.append(t_dmrg)
            t_mean = np.mean(time_param)
            t_std = np.std(time_param)
            new_timeout_secs = t_mean + 3*t_std
            energies_param.append(energy)
            entropies_param.append(entropy)
            schmidt_vals_param.append(schmidt_vals)
            time_param.append(t_dmrg)

    return energies_param, entropies_param, schmidt_vals_param, time_param



import numpy as np
import argparse
from qs_mps.sparse_hamiltonians_and_operators import exact_evolution_sparse
from qs_mps.utils import access_txt

parser = argparse.ArgumentParser(prog="Exact Ground State and Time Evolution - Ising")
parser.add_argument("L", help="Spin chain length", type=int)
parser.add_argument(
    "trotter_steps",
    help="It will give you how many steps you need to reach time t",
    type=int,
)
parser.add_argument(
    "h_ev", help="It will give you the magnitude of the quench", type=float
)
parser.add_argument(
    "-f", "--flip", help="Flip the middle site or not", action="store_true"
)
parser.add_argument(
    "-m", "--model", help="Model to simulate", default="Ising", type=str
)
parser.add_argument(
    "-t", "--time", help="Final time of the evolution", default=10, type=float
)
parser.add_argument(
    "-h_ti",
    "--h_transverse_init",
    help="Initial transverse field before the quench",
    default=0,
    type=float,
)
parser.add_argument(
    "-w",
    "--where",
    help="Where to compute the Schmidt decomposition",
    default=-1,
    type=int,
)
parser.add_argument(
    "-b",
    "--bond",
    help="Perform a bond Schmidt decomposition, By default True",
    action="store_false",
)

args = parser.parse_args()

if args.where == -1:
    args.where = args.L // 2
elif args.where == -2:
    args.bond = False

(
    psi_new,
    mag_exact_loc,
    mag_exact_loc_X,
    mag_exact_tot,
    entropy_tot,
) = exact_evolution_sparse(
    L=args.L,
    h_t=args.h_transverse_init,
    h_ev=args.h_ev,
    time=args.time,
    trotter_steps=args.trotter_steps,
    flip=args.flip,
    where=args.where,
    bond=args.bond,
)

if args.bond == False:
    args.where = "all"

np.savetxt(
    f"D:/code/projects/0_ISING/results/exact/mag_data/mag_exact_tot_{args.model}_L_{args.L}_flip_{args.flip}_h_ev_{args.h_ev}_trotter_steps_{args.trotter_steps}_t_{args.time}",
    mag_exact_tot,
)
np.savetxt(
    f"D:/code/projects/0_ISING/results/exact/mag_data/mag_exact_loc_X_{args.model}_L_{args.L}_flip_{args.flip}_h_ev_{args.h_ev}_trotter_steps_{args.trotter_steps}_t_{args.time}",
    mag_exact_loc_X,
)
np.savetxt(
    f"D:/code/projects/0_ISING/results/exact/mag_data/mag_exact_loc_{args.model}_L_{args.L}_flip_{args.flip}_h_ev_{args.h_ev}_trotter_steps_{args.trotter_steps}_t_{args.time}",
    mag_exact_loc,
)
mag_exact_loc_Z = access_txt(
    f"D:/code/projects/0_ISING/results/exact/mag_data/mag_exact_loc_{args.model}_L_{args.L}_flip_{args.flip}_h_ev_{args.h_ev}_trotter_steps_{args.trotter_steps}_t_{args.time}",
    args.L // 2,
)
np.savetxt(
    f"D:/code/projects/0_ISING/results/exact/mag_data/mag_exact_loc_Z_{args.model}_L_{args.L}_flip_{args.flip}_h_ev_{args.h_ev}_trotter_steps_{args.trotter_steps}_t_{args.time}",
    mag_exact_loc_Z,
)
np.savetxt(
    f"D:/code/projects/0_ISING/results/exact/entropy/exact_entropy_{args.where}_{args.model}_L_{args.L}_flip_{args.flip}_h_ev_{args.h_ev}_trotter_steps_{args.trotter_steps}_t_{args.time}",
    entropy_tot,
)



"""
We reproduce the analytical formulas of the magnetization and
two-point correlation function for the 1D Transverse Field Ising 
Chain.

"""
# %%
from sympy import *
import numpy as np
import matplotlib.pyplot as plt

# %%
h, k = symbols("h k")
eps_k = sqrt(h**2 - 2 * h * cos(k) + 1)
der_eps_k = eps_k.diff(k)
# %%
h_0 = symbols("h_0")
cos_k = (h * h_0 - (h + h_0) * cos(k) + 1) / (eps_k.subs(h, h_0) * eps_k)
# %%
# I = der_eps_k*ln(abs(cos_k))/pi
I = ln(abs(cos(k))) / pi
integral = Integral(I.subs([(h, 0.3), (h_0, 0)]), (k, 0, pi))
# %%
# example
k, a = symbols("k a")
sin_func = a * sin(k)
cos_func = a * cos(k)
der_cos = diff(cos_func, k)
prod = der_cos * sin_func
I = Integral(prod.subs(a, pi), (k, 0, pi))
# %%
# integration by parts
prod = cos_func.subs(a, pi) * sin_func.subs(a, pi)
first = prod.subs(x, pi) - prod.subs(x, 0)
der_sin = diff(sin_func, x)
second = Integral(der_sin * cos_func, (x, 0, pi)).subs(a, pi)
# %%
f_der = der_eps_k
f = eps_k
g = ln(abs(cos_k))
g_der = diff(g, k)
# %%
prod = f.subs(h, 0.3) * g.subs([(h, 0.3), (h_0, 0)])
first = prod.subs(k, pi) - prod.subs(k, 0)
second = Integral(f * g_der, (k, 0, pi))
# %%
